from gettext import gettext as _

from django.db import transaction
from rest_framework import serializers
from rest_framework.exceptions import ValidationError as DRFValidationError
from rest_framework.validators import UniqueValidator

from pulpcore.app import models
from pulpcore.app.serializers import DetailIdentityField, DetailRelatedField, ModelSerializer
from pulpcore.tasking.util import get_url


class AlternateContentSourcePathField(serializers.ListField):
    """Serializer field for AlternateContentSource."""

    child = serializers.CharField()

    def to_representation(self, paths):
        """
        A serializer field for AlternateContentSourcePath models.

        Args:
            acs_pk (pk of AlternateContentSource instance): UUID of AlternateContentSource

        Returns:
            A list of paths related to AlternateContentSource
        """
        return [acs_path.path for acs_path in paths.all()]


class AlternateContentSourceSerializer(ModelSerializer):
    """
    Serializer for the AlternateContentSource.
    """

    pulp_href = DetailIdentityField(view_name_pattern=r"acs(-.*/.*)-detail")
    name = serializers.CharField(
        help_text=_("Name of Alternate Content Source."),
        required=True,
        validators=[UniqueValidator(queryset=models.AlternateContentSource.objects.all())],
    )
    last_refreshed = serializers.DateTimeField(
        help_text=_("Date of last refresh of AlternateContentSource."),
        allow_null=True,
        required=False,
    )
    remote = DetailRelatedField(
        help_text=_("The remote to provide alternate content source."),
        view_name_pattern=r"remotes(-.*/.*)-detail",
        queryset=models.Remote.objects.all(),
        required=True,
    )
    paths = AlternateContentSourcePathField(
        help_text=_(
            "List of paths that will be appended to the Remote url when searching for content."
        ),
        required=False,
    )

    def validate_remote(self, remote):
        if remote.policy != "on_demand":
            raise serializers.ValidationError(
                _("Remote used with alternate content source must have the 'on_demand' policy.")
            )

        if type(remote) not in self.Meta.model.REMOTE_TYPES:
            raise serializers.ValidationError(
                detail=_("Type for Remote '{}' does not match ACS type.").format(remote.name)
            )
        return remote

    @transaction.atomic
    def create(self, validated_data):
        """Create Alternate Content Source and its path if specified."""
        paths = validated_data.pop("paths", [])
        acs = super().create(validated_data)

        try:
            self._update_paths(acs, paths)
        except DRFValidationError as exc:
            acs.delete()
            raise exc
        return acs

    def _update_paths(self, acs, paths):
        """Update Alternate Content Source paths."""
        existing_paths = {
            acs_path.path
            for acs_path in models.AlternateContentSourcePath.objects.filter(
                alternate_content_source=acs.pk
            )
        }
        if paths is None:
            to_remove = set()
            to_add = set()
        else:
            to_remove = existing_paths - set(paths)
            to_add = set(paths) - existing_paths
        if to_remove:
            models.AlternateContentSourcePath.objects.filter(path__in=to_remove).delete()
        if to_add:
            for acs_path in to_add:
                new_path = {
                    "alternate_content_source": get_url(acs),
                    "path": acs_path,
                }
                acs_path_serializer = AlternateContentSourcePathSerializer(data=new_path)
                acs_path_serializer.is_valid(raise_exception=True)
                acs_path_serializer.save()

        # if no paths for an ACS, we need create empty path to use base path of ACS remote
        if not models.AlternateContentSourcePath.objects.filter(
            alternate_content_source=acs.pk
        ).exists():
            empty_path_serializer_data = {"alternate_content_source": get_url(acs), "path": ""}
            empty_path_serializer = AlternateContentSourcePathSerializer(
                data=empty_path_serializer_data
            )
            empty_path_serializer.is_valid(raise_exception=True)
            empty_path_serializer.save()

    def update(self, instance, validated_data):
        """Update an Alternate Content Source."""
        instance.name = validated_data.get("name", instance.name)
        instance.remote = validated_data.get("remote", instance.remote)
        paths = validated_data.get("paths")
        with transaction.atomic():
            self._update_paths(instance, paths)
            instance.save()
        return instance

    class Meta:
        model = models.AlternateContentSource
        fields = ModelSerializer.Meta.fields + (
            "pulp_href",
            "name",
            "last_refreshed",
            "paths",
            "remote",
        )


class AlternateContentSourcePathSerializer(ModelSerializer):
    """
    Serializer for the AlternateContentSourcePath.
    """

    alternate_content_source = DetailRelatedField(
        view_name_pattern=r"acs(-.*/.*)-detail",
        queryset=models.AlternateContentSource.objects.all(),
        required=True,
    )
    path = serializers.CharField(help_text=_("Path for ACS."), allow_blank=True, required=False)
    repository = DetailRelatedField(
        view_name_pattern=r"repository(-.*/.*)-detail",
        queryset=models.Repository.objects.all(),
        required=False,
        allow_null=True,
    )

    class Meta:
        model = models.AlternateContentSourcePath
        fields = ModelSerializer.Meta.fields + ("alternate_content_source", "path", "repository")



"""
Tests PulpImporter and PulpImport functionality

NOTE: assumes ALLOWED_EXPORT_PATHS and ALLOWED_IMPORT_PATHS settings contain "/tmp" - all tests
will fail if this is not the case.
"""
import json
import unittest

from pulp_smash import api, cli, config
from pulp_smash.utils import uuid4, get_pulp_setting
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, monitor_task_group
from pulp_smash.pulp3.utils import (
    gen_repo,
)

from pulpcore.tests.functional.api.using_plugin.utils import (
    create_repo_and_versions,
    delete_exporter,
    gen_file_client,
    gen_file_remote,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ExportersPulpExportsApi,
    ExportersPulpApi,
    ImportersPulpImportCheckApi,
    ImportersPulpImportsApi,
    ImportersPulpApi,
)

from pulpcore.client.pulpcore.exceptions import ApiException

from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
    RepositorySyncURL,
    RemotesFileApi,
)

NUM_REPOS = 2


class PulpImportTestCase(unittest.TestCase):
    """
    Base functionality for PulpImporter and PulpImport test classes
    """

    @classmethod
    def _setup_repositories(cls):
        """Create and sync a number of repositories to be exported."""
        # create and remember a set of repo
        import_repos = []
        export_repos = []
        remotes = []
        for r in range(NUM_REPOS):
            import_repo = cls.repo_api.create(gen_repo())
            export_repo = cls.repo_api.create(gen_repo())
            body = gen_file_remote()
            remote = cls.remote_api.create(body)
            repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
            sync_response = cls.repo_api.sync(export_repo.pulp_href, repository_sync_data)
            monitor_task(sync_response.task)
            export_repo = cls.repo_api.read(export_repo.pulp_href)
            # remember it
            export_repos.append(export_repo)
            import_repos.append(import_repo)
            remotes.append(remote)
        return import_repos, export_repos, remotes

    @classmethod
    def _create_exporter(cls, cleanup=True):
        body = {
            "name": uuid4(),
            "repositories": [r.pulp_href for r in cls.export_repos],
            "path": "/tmp/{}".format(uuid4()),
        }
        exporter = cls.exporter_api.create(body)
        return exporter

    @classmethod
    def _create_export(cls):
        export_response = cls.exports_api.create(cls.exporter.pulp_href, {})
        monitor_task(export_response.task)
        task = cls.client.get(export_response.task)
        resources = task["created_resources"]
        export_href = resources[0]
        export = cls.exports_api.read(export_href)
        return export

    @classmethod
    def _create_chunked_export(cls):
        export_response = cls.exports_api.create(cls.exporter.pulp_href, {"chunk_size": "5KB"})
        monitor_task(export_response.task)
        task = cls.client.get(export_response.task)
        resources = task["created_resources"]
        export_href = resources[0]
        export = cls.exports_api.read(export_href)
        return export

    @classmethod
    def _setup_import_check_directories(cls):
        """Creates a directory/file structure for testing import-check"""
        cli_client = cli.Client(cls.cfg)
        cmd = (
            "mkdir",
            "-p",
            "/tmp/importcheck/noreaddir",
            "/tmp/importcheck/nowritedir",
            "/tmp/importcheck/nowritedir/notafile",
        )
        cli_client.run(cmd, sudo=False)

        cmd = ("touch", "/tmp/importcheck/noreadfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("touch", "/tmp/importcheck/noreaddir/goodfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("touch", "/tmp/importcheck/nowritedir/goodfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("touch", "/tmp/importcheck/nowritedir/noreadfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("chmod", "333", "/tmp/importcheck/nowritedir/noreadfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("chmod", "333", "/tmp/importcheck/noreadfile")
        cli_client.run(cmd, sudo=False)

        cmd = ("chmod", "333", "/tmp/importcheck/noreaddir")
        cli_client.run(cmd, sudo=False)

        cmd = ("chmod", "555", "/tmp/importcheck/nowritedir")
        cli_client.run(cmd, sudo=False)

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.cli_client = cli.Client(cls.cfg)
        allowed_imports = get_pulp_setting(cls.cli_client, "ALLOWED_IMPORT_PATHS")
        if not allowed_imports or "/tmp" not in allowed_imports:
            raise unittest.SkipTest(
                "Cannot run import-tests unless /tmp is in ALLOWED_IMPORT_PATHS ({}).".format(
                    allowed_imports
                ),
            )

        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.core_client = CoreApiClient(configuration=cls.cfg.get_bindings_config())
        cls.file_client = gen_file_client()

        cls.repo_api = RepositoriesFileApi(cls.file_client)
        cls.remote_api = RemotesFileApi(cls.file_client)
        cls.versions_api = RepositoriesFileVersionsApi(cls.file_client)
        cls.content_api = ContentFilesApi(cls.file_client)
        cls.exporter_api = ExportersPulpApi(cls.core_client)
        cls.exports_api = ExportersPulpExportsApi(cls.core_client)
        cls.importer_api = ImportersPulpApi(cls.core_client)
        cls.imports_api = ImportersPulpImportsApi(cls.core_client)

        cls.import_check_api = ImportersPulpImportCheckApi(cls.core_client)

        (cls.import_repos, cls.export_repos, cls.remotes) = cls._setup_repositories()
        cls.exporter = cls._create_exporter()
        cls.export = cls._create_export()
        cls.chunked_export = cls._create_chunked_export()
        cls._setup_import_check_directories()

    @classmethod
    def _delete_import_check_structures(cls):
        """Deletes the directory tree used for testing import-check"""
        cli_client = cli.Client(cls.cfg)
        cmd = ("chmod", "-R", "+rwx", "/tmp/importcheck/")
        cli_client.run(cmd, sudo=False)
        cmd = ("rm", "-rf", "/tmp/importcheck")
        cli_client.run(cmd, sudo=False)

    @classmethod
    def _create_repo_and_versions(cls):
        a_repo, versions = create_repo_and_versions(
            cls.export_repos[0], cls.repo_api, cls.versions_api, cls.content_api
        )
        return a_repo, versions

    @classmethod
    def tearDownClass(cls):
        """Clean up."""
        for remote in cls.remotes:
            cls.remote_api.delete(remote.pulp_href)
        for repo in cls.export_repos:
            cls.repo_api.delete(repo.pulp_href)
        for repo in cls.import_repos:
            cls.repo_api.delete(repo.pulp_href)
        delete_exporter(cls.exporter)
        cls._delete_import_check_structures()
        delete_orphans()

    def _create_importer(self, name=None, cleanup=True, exported_repos=None):
        """Create an importer."""
        mapping = {}
        if not name:
            name = uuid4()
        if not exported_repos:
            exported_repos = self.export_repos

        for idx, repo in enumerate(exported_repos):
            mapping[repo.name] = self.import_repos[idx].name

        body = {
            "name": name,
            "repo_mapping": mapping,
        }

        importer = self.importer_api.create(body)

        if cleanup:
            self.addCleanup(self.importer_api.delete, importer.pulp_href)

        return importer

    def _find_toc(self):
        filenames = [
            f for f in list(self.chunked_export.output_file_info.keys()) if f.endswith("json")
        ]
        return filenames[0]

    def _find_path(self):
        filenames = [f for f in list(self.export.output_file_info.keys()) if f.endswith("tar.gz")]
        return filenames[0]

    def _perform_import(self, importer, chunked=False, an_export=None):
        """Perform an import with importer."""
        if not an_export:
            an_export = self.chunked_export if chunked else self.export

        if chunked:
            filenames = [f for f in list(an_export.output_file_info.keys()) if f.endswith("json")]
            import_response = self.imports_api.create(importer.pulp_href, {"toc": filenames[0]})
        else:
            filenames = [f for f in list(an_export.output_file_info.keys()) if f.endswith("tar.gz")]
            import_response = self.imports_api.create(importer.pulp_href, {"path": filenames[0]})
        task_group = monitor_task_group(import_response.task_group)

        return task_group

    def test_importer_create(self):
        """Test creating an importer."""
        name = uuid4()
        importer = self._create_importer(name)

        self.assertEqual(importer.name, name)
        importer = self.importer_api.read(importer.pulp_href)
        self.assertEqual(importer.name, name)

    def test_importer_delete(self):
        """Test deleting an importer."""
        importer = self._create_importer(cleanup=False)

        self.importer_api.delete(importer.pulp_href)

        with self.assertRaises(ApiException) as ae:
            self.importer_api.read(importer.pulp_href)

        self.assertEqual(404, ae.exception.status)

    def test_import(self):
        """Test an import."""
        importer = self._create_importer()
        task_group = self._perform_import(importer)
        self.assertEqual(len(self.import_repos) + 1, task_group.completed)

        for report in task_group.group_progress_reports:
            if report.code == "import.repo.versions":
                self.assertEqual(report.done, len(self.import_repos))

        for repo in self.import_repos:
            repo = self.repo_api.read(repo.pulp_href)
            self.assertEqual(f"{repo.pulp_href}versions/1/", repo.latest_version_href)

    def test_double_import(self):
        """Test two imports of our export."""
        importer = self._create_importer()
        self._perform_import(importer)
        self._perform_import(importer)

        imports = self.imports_api.list(importer.pulp_href).results
        self.assertEqual(len(imports), 2)

        for repo in self.import_repos:
            repo = self.repo_api.read(repo.pulp_href)
            # still only one version as pulp won't create a new version if nothing changed
            self.assertEqual(f"{repo.pulp_href}versions/1/", repo.latest_version_href)

    def test_chunked_import(self):
        """Test an import."""
        importer = self._create_importer()
        task_group = self._perform_import(importer, chunked=True)
        self.assertEqual(len(self.import_repos) + 1, task_group.completed)
        for repo in self.import_repos:
            repo = self.repo_api.read(repo.pulp_href)
            self.assertEqual(f"{repo.pulp_href}versions/1/", repo.latest_version_href)

    def test_import_check_valid_path(self):
        body = {"path": self._find_path()}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.path.context, self._find_path())
        self.assertTrue(result.path.is_valid)
        self.assertEqual(len(result.path.messages), 0)
        self.assertIsNone(result.toc)
        self.assertIsNone(result.repo_mapping)

    def test_import_check_valid_toc(self):
        body = {"toc": self._find_toc()}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.toc.context, self._find_toc())
        self.assertTrue(result.toc.is_valid)
        self.assertEqual(len(result.toc.messages), 0)
        self.assertIsNone(result.path)
        self.assertIsNone(result.repo_mapping)

    def test_import_check_repo_mapping(self):
        body = {"repo_mapping": json.dumps({"foo": "bar"})}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.repo_mapping.context, json.dumps({"foo": "bar"}))
        self.assertTrue(result.repo_mapping.is_valid)
        self.assertEqual(len(result.repo_mapping.messages), 0)
        self.assertIsNone(result.path)
        self.assertIsNone(result.toc)

        body = {"repo_mapping": '{"foo": "bar"'}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.repo_mapping.context, '{"foo": "bar"')
        self.assertFalse(result.repo_mapping.is_valid)
        self.assertEqual(result.repo_mapping.messages[0], "invalid JSON")

    def test_import_check_not_allowed(self):
        body = {"path": "/notinallowedimports"}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.path.context, "/notinallowedimports")
        self.assertFalse(result.path.is_valid)
        self.assertEqual(len(result.path.messages), 1, "Only not-allowed should be returned")
        self.assertEqual(result.path.messages[0], "/ is not an allowed import path")

        body = {"toc": "/notinallowedimports"}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.toc.context, "/notinallowedimports")
        self.assertFalse(result.toc.is_valid)
        self.assertEqual(len(result.toc.messages), 1, "Only not-allowed should be returned")
        self.assertEqual(result.toc.messages[0], "/ is not an allowed import path")

    def test_import_check_no_file(self):
        body = {"path": "/tmp/idonotexist"}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.path.context, "/tmp/idonotexist")
        self.assertFalse(result.path.is_valid)
        self.assertTrue(
            any("file /tmp/idonotexist does not exist" in s for s in result.path.messages)
        )

        body = {"toc": "/tmp/idonotexist"}
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.toc.context, "/tmp/idonotexist")
        self.assertFalse(result.toc.is_valid)
        self.assertTrue(
            any("file /tmp/idonotexist does not exist" in s for s in result.toc.messages)
        )

    def test_import_check_all_valid(self):
        body = {
            "path": self._find_path(),
            "toc": self._find_toc(),
            "repo_mapping": json.dumps({"foo": "bar"}),
        }
        result = self.import_check_api.pulp_import_check_post(body)
        self.assertEqual(result.path.context, self._find_path())
        self.assertEqual(result.toc.context, self._find_toc())
        self.assertEqual(result.repo_mapping.context, json.dumps({"foo": "bar"}))

        self.assertTrue(result.path.is_valid)
        self.assertTrue(result.toc.is_valid)
        self.assertTrue(result.repo_mapping.is_valid)

        self.assertEqual(len(result.path.messages), 0)
        self.assertEqual(len(result.toc.messages), 0)
        self.assertEqual(len(result.repo_mapping.messages), 0)

    def test_import_check_multiple_errors(self):
        body = {
            "path": "/notinallowedimports",
            "toc": "/tmp/importcheck/nowritedir/notafile",
            "repo_mapping": '{"foo": "bar"',
        }
        result = self.import_check_api.pulp_import_check_post(body)

        self.assertFalse(result.path.is_valid)
        self.assertEqual(len(result.path.messages), 1, "Only not-allowed should be returned")
        self.assertEqual(result.path.messages[0], "/ is not an allowed import path")

        self.assertFalse(result.toc.is_valid)
        self.assertTrue(
            any(
                "/tmp/importcheck/nowritedir/notafile is not a file" in s
                for s in result.toc.messages
            )
        )
        # FAILS IN CI, passes locally
        # self.assertTrue(
        #     any(
        #         "directory /tmp/importcheck/nowritedir must allow pulp write-access" in s
        #         for s in result.toc.messages
        #     )
        # )

        self.assertFalse(result.repo_mapping.is_valid)
        self.assertEqual(result.repo_mapping.messages[0], "invalid JSON")

    def _gen_export(self, exporter, body={}):
        """Create and read back an export for the specified PulpExporter."""
        export_response = self.exports_api.create(exporter.pulp_href, body)
        monitor_task(export_response.task)
        task = self.client.get(export_response.task)
        resources = task["created_resources"]
        export_href = resources[0]
        export = self.exports_api.read(export_href)
        return export

    def _export_first_version(self, a_repo, versions):
        body = {
            "name": uuid4(),
            "repositories": [a_repo.pulp_href],
            "path": "/tmp/{}".format(uuid4()),
        }
        exporter = self.exporter_api.create(body)
        self.addCleanup(delete_exporter, exporter)
        # export from version-0 to version-1, last=v1
        body = {
            "start_versions": [versions.results[0].pulp_href],
            "versions": [versions.results[1].pulp_href],
            "full": False,
        }
        export = self._gen_export(exporter, body)
        return export

    def test_import_not_latest_version(self):
        try:
            repo, versions = self._create_repo_and_versions()

            export = self._export_first_version(repo, versions)
            """Test an import."""
            importer = self._create_importer(exported_repos=[repo])
            task_group = self._perform_import(importer, chunked=False, an_export=export)

            for report in task_group.group_progress_reports:
                if report.code == "import.repo.versions":
                    self.assertEqual(report.done, 1)

            imported_repo = self.repo_api.read(self.import_repos[0].pulp_href)
            self.assertNotEqual(
                f"{imported_repo.pulp_href}versions/0/", imported_repo.latest_version_href
            )
        finally:
            self.repo_api.delete(repo.pulp_href)



from pathlib import Path

from django.conf import settings
from django.core.checks import Warning as CheckWarning, register


@register(deploy=True)
def storage_paths(app_configs, **kwargs):
    warnings = []

    if settings.DEFAULT_FILE_STORAGE == "pulpcore.app.models.storage.FileSystem":
        try:
            media_root_dev = Path(settings.MEDIA_ROOT).stat().st_dev
        except OSError:
            media_root_dev = None
            warnings.append(
                CheckWarning(
                    "Your MEDIA_ROOT setting points to a path that does not exist.",
                    id="pulpcore.W001",
                )
            )

        try:
            working_dir_dev = Path(settings.WORKING_DIRECTORY).stat().st_dev
        except OSError:
            working_dir_dev = None
            warnings.append(
                CheckWarning(
                    "Your WORKING_DIRECTORY setting points to a path that does not exist.",
                    id="pulpcore.W002",
                )
            )

        if media_root_dev and media_root_dev != working_dir_dev:
            warnings.append(
                CheckWarning(
                    "MEDIA_ROOT and WORKING_DIRECTORY are on different filesystems. "
                    "It is highly recommended that these live on the same filesystem",
                    id="pulpcore.W003",
                )
            )

    return warnings



import unittest

from django.http import Http404, QueryDict
from django.test import TestCase

from pulpcore.app import models, serializers, viewsets


class TestGetQuerySet(TestCase):
    @unittest.skip("fails for unknown reasons")
    def test_adds_filters(self):
        """
        Tests to make sure the correct lookup is being added to the queryset based on its
        'parent_lookup_kwargs' value.
        """
        repo = models.Repository.objects.create(name="foo")
        repo2 = models.Repository.objects.create(name="foo2")
        # no concurrency so this is fine
        models.RepositoryVersion.objects.create(repository=repo, number=1)
        models.RepositoryVersion.objects.create(repository=repo2, number=1)
        viewset = viewsets.RepositoryVersionViewSet()
        viewset.kwargs = {"repository_pk": repo.pk}
        queryset = viewset.get_queryset()
        expected = models.RepositoryVersion.objects.filter(repository__pk=repo.pk)

        # weird, stupid django quirk
        # https://docs.djangoproject.com/en/2.0/topics/testing/tools/#django.test.TransactionTestCase.assertQuerysetEqual
        self.assertQuerysetEqual(queryset, map(repr, expected))

    def test_does_not_add_filters(self):
        """
        Tests to make sure no filters are applied, based on its empty 'parent_lookup_kwargs'
        value.
        """
        models.Repository.objects.create(name="foo")
        viewset = viewsets.RepositoryViewSet()
        viewset.kwargs = {"name": "foo"}
        queryset = viewset.get_queryset()
        expected = models.Repository.objects.all()

        # weird, stupid django quirk
        # https://docs.djangoproject.com/en/2.0/topics/testing/tools/#django.test.TransactionTestCase.assertQuerysetEqual
        self.assertQuerysetEqual(queryset, map(repr, expected))


class TestGetSerializerClass(TestCase):
    def test_must_define_serializer_class(self):
        """
        Test that get_serializer_class() raises an AssertionError if you don't define the
        serializer_class attribute.
        """

        class TestTaskViewSet(viewsets.NamedModelViewSet):
            minimal_serializer_class = serializers.MinimalTaskSerializer

        with self.assertRaises(AssertionError):
            TestTaskViewSet().get_serializer_class()

    def test_serializer_class(self):
        """
        Tests that get_serializer_class() returns the serializer_class attribute if it exists,
        and that it doesn't error if no minimal serializer is defined, but minimal=True.
        """

        class TestTaskViewSet(viewsets.NamedModelViewSet):
            serializer_class = serializers.TaskSerializer

        viewset = TestTaskViewSet()
        self.assertEquals(viewset.get_serializer_class(), serializers.TaskSerializer)

        request = unittest.mock.MagicMock()
        request.query_params = QueryDict("minimal=True")
        viewset.request = request

        self.assertEquals(viewset.get_serializer_class(), serializers.TaskSerializer)

    def test_minimal_query_param(self):
        """
        Tests that get_serializer_class() returns the correct serializer in the correct situations.
        """

        class TestTaskViewSet(viewsets.NamedModelViewSet):
            serializer_class = serializers.TaskSerializer
            minimal_serializer_class = serializers.MinimalTaskSerializer

        viewset = TestTaskViewSet()
        request = unittest.mock.MagicMock()

        # Test that it uses the full serializer with no query params
        request.query_params = QueryDict()
        viewset.request = request
        self.assertEquals(viewset.get_serializer_class(), serializers.TaskSerializer)
        # Test that it uses the full serializer with minimal=False
        request.query_params = QueryDict("minimal=False")
        viewset.request = request
        self.assertEquals(viewset.get_serializer_class(), serializers.TaskSerializer)
        # Test that it uses the minimal serializer with minimal=True
        request.query_params = QueryDict("minimal=True")
        viewset.request = request
        self.assertEquals(viewset.get_serializer_class(), serializers.MinimalTaskSerializer)


class TestGetParentFieldAndObject(TestCase):
    def test_no_parent_object(self):
        """
        Tests that get_parent_field_and_object() raises django.http.Http404 if the parent object
        does not exist on a nested viewset.
        """
        viewset = viewsets.RepositoryVersionViewSet()
        viewset.kwargs = {"repository_pk": 500}

        with self.assertRaises(Http404):
            viewset.get_parent_field_and_object()

    def test_get_parent_field_and_object(self):
        """
        Tests that get_parent_field_and_object() returns the correct parent field and parent
        object.
        """
        repo = models.Repository.objects.create(name="foo")
        viewset = viewsets.RepositoryVersionViewSet()
        viewset.kwargs = {"repository_pk": repo.pk}

        self.assertEquals(("repository", repo), viewset.get_parent_field_and_object())

    def test_get_parent_object(self):
        """
        Tests that get_parent_object() returns the correct parent object.
        """
        repo = models.Repository.objects.create(name="foo")
        viewset = viewsets.RepositoryVersionViewSet()
        viewset.kwargs = {"repository_pk": repo.pk}

        self.assertEquals(repo, viewset.get_parent_object())


class TestGetNestDepth(TestCase):
    def test_get_nest_depth(self):
        """
        Test that _get_nest_depth() returns the correct nesting depths.
        """
        self.assertEquals(1, viewsets.RepositoryViewSet._get_nest_depth())
        self.assertEquals(2, viewsets.RepositoryVersionViewSet._get_nest_depth())



from gettext import gettext as _
import json
import os
from drf_spectacular.utils import extend_schema
from rest_framework.views import APIView
from rest_framework.response import Response

from pulpcore.app import settings
from pulpcore.app.serializers import PulpImportCheckResponseSerializer, PulpImportCheckSerializer


def _check_allowed_import_path(a_path):
    user_provided_realpath = os.path.realpath(a_path)
    for allowed_path in settings.ALLOWED_IMPORT_PATHS:
        if user_provided_realpath.startswith(allowed_path):
            return True, None
    return False, _(
        "{} is not an allowed import path".format(os.path.dirname(os.path.realpath(a_path)))
    )


def _validate_file(in_param, data):
    """
    Returns a (is-valid, msgs[]) tuple describing all problems found with data[in_param]

    We check for a number of things, attempting to return all the errors we can find. We don't want
    to give out information for files in arbitrary locations on the filesystem; if the check
    for ALLOWED_IMPORT_PATHS fails, we report that and ignore any other problems.

    If the directory containing the base-file doesn't exist, or isn't readable, or the specified
    file doesn't exist, report and return.

    Error-messages for all other checks are additive.
    """
    # check allowed, leave if failed
    file = data[in_param]
    real_file = os.path.realpath(file)
    rc, msg = _check_allowed_import_path(real_file)
    if not rc:
        return rc, [msg]

    # check directory-sanity, leave if failed
    owning_dir = os.path.dirname(real_file)
    if not os.path.exists(owning_dir):
        return False, [_("directory {} does not exist").format(owning_dir)]
    if not os.access(owning_dir, os.R_OK):
        return False, [_("directory {} does not allow read-access").format(owning_dir)]

    # check file-exists, leave if failed
    if not os.path.exists(real_file):
        return False, [_("file {} does not exist").format(real_file)]

    # check file-sanity
    msgs = []
    isfile = os.path.isfile(real_file)
    readable = os.access(real_file, os.R_OK)

    rc = isfile and readable
    if not isfile:
        msgs.append(_("{} is not a file".format(real_file)))
    if not readable:
        msgs.append(_("{} exists but cannot be read".format(real_file)))

    # extra check for toc-dir-write
    if in_param == "toc":
        if not os.access(owning_dir, os.W_OK):
            rc = False
            msgs.append(_("directory {} must allow pulp write-access".format(owning_dir)))

    return rc, msgs


class PulpImporterImportCheckView(APIView):
    """
    Returns validity of proposed parameters for a PulpImport call.
    """

    @extend_schema(
        summary="Validate the parameters to be used for a PulpImport call",
        operation_id="pulp_import_check_post",
        request=PulpImportCheckSerializer,
        responses={200: PulpImportCheckResponseSerializer},
    )
    def post(self, request, format=None):
        """
        Evaluates validity of proposed PulpImport parameters 'toc', 'path', and 'repo_mapping'.

        * Checks that toc, path are in ALLOWED_IMPORT_PATHS
        * if ALLOWED:
          * Checks that toc, path exist and are readable
          * If toc specified, checks that containing dir is writeable
        * Checks that repo_mapping is valid JSON
        """
        serializer = PulpImportCheckSerializer(data=request.data)
        if serializer.is_valid():
            data = {}
            if "toc" in serializer.data:
                data["toc"] = {}
                data["toc"]["context"] = serializer.data["toc"]
                data["toc"]["is_valid"], data["toc"]["messages"] = _validate_file(
                    "toc", serializer.data
                )

            if "path" in serializer.data:
                data["path"] = {}
                data["path"]["context"] = serializer.data["path"]
                data["path"]["is_valid"], data["path"]["messages"] = _validate_file(
                    "path", serializer.data
                )

            if "repo_mapping" in serializer.data:
                data["repo_mapping"] = {}
                data["repo_mapping"]["context"] = serializer.data["repo_mapping"]
                try:
                    json.loads(serializer.data["repo_mapping"])
                    data["repo_mapping"]["is_valid"] = True
                    data["repo_mapping"]["messages"] = []
                except json.JSONDecodeError:
                    data["repo_mapping"]["is_valid"] = False
                    data["repo_mapping"]["messages"] = [_("invalid JSON")]

            crs = PulpImportCheckResponseSerializer(data, context={"request": request})
            return Response(crs.data)
        return Response(serializer.errors, status=400)



# Generated by Django 2.2.19 on 2021-04-01 14:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0063_repository_retained_versions'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='user_hidden',
            field=models.BooleanField(default=False),
        ),
    ]



from gettext import gettext as _

from django.core.management import BaseCommand


class Command(BaseCommand):
    """
    Django management command for printing a summary report of a Stages API pipeline run.
    """

    help = _(
        "Print a summary of a Stages API pipeline run. This command is provided as a tech "
        "preview and may not work properly or change in the future."
    )

    def add_arguments(self, parser):
        parser.add_argument("file_path", help=_("The path to the sqlite3 db with the run data."))

    def handle(self, *args, **options):
        import sqlite3

        CONN = sqlite3.connect(options["file_path"])
        c = CONN.cursor()

        c.execute("SELECT uuid, name, num FROM stages ORDER BY num ASC")

        stages = []
        stages_map = {}
        for row in c.fetchall():
            new_dict = {
                "uuid": row[0],
                "name": row[1],
                "waiting_time_sum": 0.0,
                "service_time_sum": 0.0,
                "length_sum": 0,
                "interarrival_time_sum": 0.0,
            }
            stages.append(new_dict)
            stages_map[row[0]] = new_dict

        c.execute("SELECT uuid, waiting_time, service_time FROM traffic")

        in_queue_count = 0
        for row in c.fetchall():
            in_queue_count = in_queue_count + 1
            stages_map[row[0]]["waiting_time_sum"] += row[1]
            stages_map[row[0]]["service_time_sum"] += row[2]

        c.execute("SELECT uuid, length, interarrival_time FROM system")

        arrival_count = 0
        for row in c.fetchall():
            arrival_count = arrival_count + 1
            stages_map[row[0]]["length_sum"] += row[1]
            stages_map[row[0]]["interarrival_time_sum"] += row[2]

        for stage in stages:
            msg = _("{name}\n\tservice time average: {srv_avg:4f}\n")
            if in_queue_count == 0:
                # This is the first queue that gets no put() calls, so avoid DivisionByZero errors
                waiting_avg = srv_avg = length_avg = interarrival_avg = 0
            else:
                waiting_avg = stage["waiting_time_sum"] / in_queue_count
                srv_avg = stage["service_time_sum"] / in_queue_count
                length_avg = stage["length_sum"] / arrival_count
                interarrival_avg = stage["interarrival_time_sum"] / arrival_count

            print(
                "\n"
                "    |\n"
                "    |waiting time average: {wt:4f}\n"
                "    |queue length average: {ln:4f}\n"
                "    |interarrival average: {inter:4f}\n"
                "    |\n"
                "    \u030C\n".format(wt=waiting_avg, ln=length_avg, inter=interarrival_avg)
            )
            print(msg.format(name=stage["name"], srv_avg=srv_avg))



import click
import logging
import os

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "pulpcore.app.settings")

django.setup()

from pulpcore.tasking.pulpcore_worker import NewPulpWorker  # noqa: E402: module level not at top


_logger = logging.getLogger(__name__)


@click.option("--pid", help="Write the process ID number to a file at the specified path")
@click.command()
def worker(pid):
    """A Pulp worker."""

    if pid:
        with open(os.path.expanduser(pid), "w") as fp:
            fp.write(str(os.getpid()))

    _logger.info("Starting distributed type worker")

    NewPulpWorker().run_forever()



from gettext import gettext as _

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.auth.hashers import make_password
from django.contrib.auth.password_validation import validate_password
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from guardian.models.models import GroupObjectPermission
from rest_framework import serializers
from rest_framework.validators import UniqueValidator
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from pulpcore.app.models import Group
from pulpcore.app.models.role import GroupRole, Role, UserRole
from pulpcore.app.serializers import (
    NestedIdentityField,
    IdentityField,
    ValidateFieldsMixin,
    ModelSerializer,
)
from pulpcore.app.util import get_viewset_for_model, get_request_without_query_params

User = get_user_model()


class ThisPermissionField(serializers.CharField):
    """Read only field that represents the Permission of the object.
    The object can be either a plain Permission, or GroupObjectPermission.
    """

    def to_representation(self, value):
        permission = getattr(value, "permission", value)
        return f"{permission.content_type.app_label}.{permission.codename}"


class PermissionField(serializers.RelatedField):
    """Read write Permission field."""

    queryset = Permission.objects.all()

    def to_representation(self, value):
        permission = getattr(value, "permission", value)
        return f"{permission.content_type.app_label}.{permission.codename}"

    def to_internal_value(self, data):
        try:
            app_label, codename = data.split(".", maxsplit=1)
        except Exception:
            raise serializers.ValidationError(
                _("Permission '{name}' is not in the format '<app_label>.<codename>'.").format(
                    name=data
                )
            )
        try:
            queryset = self.get_queryset()
            return queryset.get(content_type__app_label=app_label, codename=codename)
        except Permission.DoesNotExist:
            raise serializers.ValidationError(
                _("Permission '{name}' does not exist.").format(name=data)
            )


class ContentObjectField(serializers.CharField):
    """Content object field"""

    def to_representation(self, obj):
        content_object = getattr(obj, "content_object", None)
        if content_object:
            viewset = get_viewset_for_model(obj.content_object)

            request = get_request_without_query_params(self.context)

            serializer = viewset.serializer_class(obj.content_object, context={"request": request})
            return serializer.data.get("pulp_href")

    def to_internal_value(self, data):
        # ... circular import ...
        from pulpcore.app.viewsets.base import NamedModelViewSet

        if data is None:
            return {"content_object": None}
        try:
            obj = NamedModelViewSet.get_resource(data)
        except serializers.ValidationError:
            raise serializers.ValidationError(_("Invalid value: {}.").format(data))
        return {"content_object": obj}


class PermissionSerializer(serializers.Serializer):
    """Serializer for User/Group object permission."""

    pulp_href = serializers.SerializerMethodField(read_only=True)
    id = serializers.SerializerMethodField(read_only=True)
    permission = ThisPermissionField(source="*", read_only=True)
    obj = ContentObjectField(
        help_text=_("pulp_href of the object the permission is to be asserted on."),
        source="*",
        read_only=True,
    )

    def get_id(self, obj) -> int:
        """Get model/object permission id."""
        return obj.id

    def get_pulp_href(self, obj) -> str:
        """Get model/object permission pulp_href."""
        group_pk = self.context.get("group_pk")

        if group_pk and isinstance(obj, Permission):
            return reverse("model_permissions-detail", args=[group_pk, obj.pk])

        if group_pk and isinstance(obj, GroupObjectPermission):
            return reverse("object_permissions-detail", args=[group_pk, obj.pk])

    def to_representation(self, obj):
        representation = super().to_representation(obj)

        if not self.context.get("group_pk"):
            representation.pop("pulp_href")

        return representation


class UserGroupSerializer(serializers.ModelSerializer):
    """Serializer for Groups that belong to an User."""

    name = serializers.CharField(help_text=_("Name."), max_length=150)
    pulp_href = IdentityField(view_name="groups-detail")

    class Meta:
        model = Group
        fields = ("name", "pulp_href")


class UserSerializer(serializers.ModelSerializer):
    """Serializer for User."""

    pulp_href = IdentityField(view_name="users-detail")
    id = serializers.IntegerField(read_only=True)
    username = serializers.CharField(
        help_text=_("Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."),
        max_length=150,
        validators=[UniqueValidator(queryset=User.objects.all())],
    )
    password = serializers.CharField(
        help_text=_("Users password. Set to ``null`` to disable password authentication."),
        write_only=True,
        allow_null=True,
        default=None,
        style={"input_type": "password"},
    )
    first_name = serializers.CharField(
        help_text=_("First name"), max_length=150, allow_blank=True, required=False
    )
    last_name = serializers.CharField(
        help_text=_("Last name"), max_length=150, allow_blank=True, required=False
    )
    email = serializers.EmailField(help_text=_("Email address"), allow_blank=True, required=False)
    is_staff = serializers.BooleanField(
        help_text=_("Designates whether the user can log into this admin site."), default=False
    )
    is_active = serializers.BooleanField(
        help_text=_("Designates whether this user should be treated as active."), default=True
    )
    date_joined = serializers.DateTimeField(help_text=_("Date joined"), read_only=True)
    groups = UserGroupSerializer(read_only=True, many=True)

    def validate(self, data):
        data = super().validate(data)
        if "password" in data:
            # `None` will automatically result in an unusable password
            if data["password"] is not None:
                validate_password(data["password"])
            data["password"] = make_password(data["password"])
        return data

    class Meta:
        model = User
        fields = (
            "pulp_href",
            "id",
            "username",
            "password",
            "first_name",
            "last_name",
            "email",
            "is_staff",
            "is_active",
            "date_joined",
            "groups",
        )


class GroupUserSerializer(ValidateFieldsMixin, serializers.ModelSerializer):
    """Serializer for Users that belong to a Group."""

    username = serializers.CharField(
        help_text=_("Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."),
        max_length=150,
    )
    pulp_href = IdentityField(view_name="users-detail")

    class Meta:
        model = User
        fields = ("username", "pulp_href")


class GroupSerializer(ValidateFieldsMixin, serializers.ModelSerializer):
    """Serializer for Group."""

    pulp_href = IdentityField(view_name="groups-detail")
    id = serializers.IntegerField(read_only=True)
    name = serializers.CharField(
        help_text=_("Name"),
        max_length=150,
        validators=[UniqueValidator(queryset=Group.objects.all())],
    )

    class Meta:
        model = Group
        fields = ("name", "pulp_href", "id")


class RoleSerializer(ModelSerializer):
    """Serializer for Role."""

    pulp_href = IdentityField(view_name="roles-detail")

    name = serializers.CharField(
        help_text=_("The name of this role."),
        validators=[UniqueValidator(queryset=Role.objects.all())],
    )

    description = serializers.CharField(
        help_text=_("An optional description."), required=False, allow_null=True
    )

    permissions = PermissionField(
        many=True,
        help_text=_("List of permissions defining the role."),
    )

    locked = serializers.BooleanField(
        help_text=_("True if the role is system managed."),
        read_only=True,
    )

    def create(self, validated_data):
        permissions = validated_data.pop("permissions")
        instance = super().create(validated_data)
        instance.permissions.set(permissions)
        return instance

    def update(self, instance, validated_data):
        permissions = validated_data.pop("permissions", None)
        instance = super().update(instance, validated_data)
        if permissions is not None:
            instance.permissions.set(permissions)
        return instance

    class Meta:
        model = Role
        fields = ModelSerializer.Meta.fields + (
            "name",
            "description",
            "permissions",
            "locked",
        )


class UserRoleSerializer(ModelSerializer, NestedHyperlinkedModelSerializer):
    """Serializer for UserRole."""

    pulp_href = NestedIdentityField(
        view_name="roles-detail", parent_lookup_kwargs={"user_pk": "user__pk"}
    )

    role = serializers.SlugRelatedField(
        slug_field="name",
        queryset=Role.objects.all(),
    )

    content_object = ContentObjectField(
        help_text=_("Optional pulp_href of the object the permissions are to be asserted on."),
        source="*",
        allow_null=True,
    )

    def validate(self, data):
        data = super().validate(data)
        data["user"] = User.objects.get(pk=self.context["request"].resolver_match.kwargs["user_pk"])

        natural_key_args = {
            "user_id": data["user"].pk,
            "role_id": data["role"].pk,
            "object_id": None,
            "content_type": None,
        }
        content_object = data["content_object"]
        if content_object:
            content_type = ContentType.objects.get_for_model(
                content_object, for_concrete_model=False
            )
            if not data["role"].permissions.filter(content_type__pk=content_type.id).exists():
                raise serializers.ValidationError(
                    _("The role '{}' does not carry any permission for that object.").format(
                        data["role"].name
                    )
                )
            natural_key_args["object_id"] = content_object.pk
            natural_key_args["content_type"] = content_type
        if self.Meta.model.objects.filter(**natural_key_args).exists():
            raise serializers.ValidationError(_("The role is already assigned to this user."))
        return data

    class Meta:
        model = UserRole
        fields = ModelSerializer.Meta.fields + ("role", "content_object")


class GroupRoleSerializer(ModelSerializer, NestedHyperlinkedModelSerializer):
    """Serializer for GroupRole."""

    pulp_href = NestedIdentityField(
        view_name="roles-detail", parent_lookup_kwargs={"group_pk": "group__pk"}
    )

    role = serializers.SlugRelatedField(
        slug_field="name",
        queryset=Role.objects.all(),
    )

    content_object = ContentObjectField(
        help_text=_("Optional pulp_href of the object the permissions are to be asserted on."),
        source="*",
        allow_null=True,
    )

    def validate(self, data):
        data = super().validate(data)
        data["group"] = Group.objects.get(
            pk=self.context["request"].resolver_match.kwargs["group_pk"]
        )

        natural_key_args = {
            "group_id": data["group"].pk,
            "role_id": data["role"].pk,
            "object_id": None,
            "content_type": None,
        }
        content_object = data["content_object"]
        if content_object:
            content_type = ContentType.objects.get_for_model(
                content_object, for_concrete_model=False
            )
            if not data["role"].permissions.filter(content_type__pk=content_type.id).exists():
                raise serializers.ValidationError(
                    _("The role '{}' does not carry any permission for that object.").format(
                        data["role"].name
                    )
                )
            natural_key_args["object_id"] = content_object.pk
            natural_key_args["content_type"] = content_type
        if self.Meta.model.objects.filter(**natural_key_args).exists():
            raise serializers.ValidationError(_("The role is already assigned to this group."))
        return data

    class Meta:
        model = GroupRole
        fields = ModelSerializer.Meta.fields + ("role", "content_object")


class NestedRoleSerializer(serializers.Serializer):
    """
    Serializer to add/remove object roles to/from users/groups.

    This is used in conjunction with ``pulpcore.app.viewsets.base.RolesMixin`` and requires the
    underlying object to be passed as ``content_object`` in the context.
    """

    users = serializers.ListField(
        default=[],
        child=serializers.SlugRelatedField(
            slug_field="username",
            queryset=User.objects.all(),
        ),
    )
    groups = serializers.ListField(
        default=[],
        child=serializers.SlugRelatedField(
            slug_field="name",
            queryset=Group.objects.all(),
        ),
    )

    role = serializers.SlugRelatedField(
        slug_field="name",
        queryset=Role.objects.all(),
        required=True,
    )

    def validate(self, data):
        data = super().validate(data)
        if "assign" in self.context:
            obj = self.context["content_object"]
            obj_type = ContentType.objects.get_for_model(obj, for_concrete_model=False)

            if not data["role"].permissions.filter(content_type__pk=obj_type.id).exists():
                raise serializers.ValidationError(
                    _("The role does not contain any permission for that object.")
                )

            self.user_role_pks = []
            for user in data["users"]:
                qs = UserRole.objects.filter(
                    content_type_id=obj_type.id, object_id=obj.pk, user=user, role=data["role"]
                )
                if self.context["assign"]:
                    if qs.exists():
                        raise serializers.ValidationError(
                            _("The role is already assigned to user '{user}'.").format(
                                user=user.username
                            )
                        )
                else:
                    if not qs.exists():
                        raise serializers.ValidationError(
                            _("The role is not assigned to user '{user}'.").format(
                                user=user.username
                            )
                        )
                    self.user_role_pks.append(qs.get().pk)

            self.group_role_pks = []
            for group in data["groups"]:
                qs = GroupRole.objects.filter(
                    content_type_id=obj_type.id, object_id=obj.pk, group=group, role=data["role"]
                )
                if self.context["assign"]:
                    if qs.exists():
                        raise serializers.ValidationError(
                            _("The role is already assigned to group '{group}'.").format(
                                group=group.name
                            )
                        )
                else:
                    if not qs.exists():
                        raise serializers.ValidationError(
                            _("The role is not assigned to group '{group}'.").format(
                                group=group.name
                            )
                        )
                    self.group_role_pks.append(qs.get().pk)
        return data



"""pulp URL Configuration"""
import logging

from django.conf import settings
from django.contrib import admin
from django.urls import path, include
from drf_spectacular.views import (
    SpectacularJSONAPIView,
    SpectacularYAMLAPIView,
    SpectacularRedocView,
)
from rest_framework_nested import routers

from pulpcore.app.apps import pulp_plugin_configs
from pulpcore.app.views import OrphansView, PulpImporterImportCheckView, RepairView, StatusView
from pulpcore.app.viewsets import (
    ListRepositoryVersionViewSet,
    OrphansCleanupViewset,
    ReclaimSpaceViewSet,
)
from pulpcore.constants import API_ROOT

log = logging.getLogger(__name__)


class ViewSetNode:
    """
    Each node is a tree that can register nested ViewSets with DRF nested routers.

    The structure of the tree becomes the url heirarchy when the ViewSets are registered.

    Example Structure:

        RootNode
        ├─ RepositoryViewSet
        │  ├─ PluginPublisherViewSet (non-nested)
        │  │  └─ PluginPublisherVDistributionViewSet
        │  ├─ AnotherPluginPublisherViewSet
        │  │  └─ AnotherPluginDistributionViewSet
        │  └─ FileRemoteViewSet
        └─ some-non-nested viewset
    """

    def __init__(self, viewset=None):
        """
        Create a new node.

        Args:
            viewset (pulpcore.app.viewsets.base.NamedModelViewSet): If provided, represent this
                viewset. If not provided, this is the root node.
        """
        self.viewset = viewset
        self.children = []

    def add_decendent(self, node):
        """
        Add a VSNode to the tree. If node is not a direct child, attempt to add the to each child.

        Args:
            node (ViewSetNode): A node that represents a viewset and its decendents.
        """
        # Master viewsets do not have endpoints, so they do not need to be registered
        if node.viewset.is_master_viewset():
            return
        # Non-nested viewsets are attached to the root node
        if not node.viewset.parent_viewset:
            self.children.append(node)
        # The node is a direct child if the child.parent_viewset is self.viewset.
        elif self.viewset and self.viewset is node.viewset.parent_viewset:
            self.children.append(node)
        else:
            for child in self.children:
                child.add_decendent(node)

    def register_with(self, router):
        """
        Register this tree with the specified router and create new routers as necessary.

        Args:
            router (routers.DefaultRouter): router to register the viewset with.
            created_routers (list): A running list of all routers.
        Returns:
            list: List of new routers, including those created recursively.
        """
        created_routers = []
        # Root node does not need to be registered, and it doesn't need a router either.
        if self.viewset:
            router.register(self.viewset.urlpattern(), self.viewset, self.viewset.view_name())
            if self.children:
                router = routers.NestedDefaultRouter(
                    router, self.viewset.urlpattern(), lookup=self.viewset.router_lookup
                )
                created_routers.append(router)
        # If we created a new router for the parent, recursively register the children with it
        for child in self.children:
            created_routers = created_routers + child.register_with(router)
        return created_routers

    def __repr__(self):
        if not self.viewset:
            return "Root"
        else:
            return str(self.viewset)


all_viewsets = []
plugin_patterns = []
# Iterate over each app, including pulpcore and the plugins.
for app_config in pulp_plugin_configs():
    for viewsets in app_config.named_viewsets.values():
        all_viewsets.extend(viewsets)
    if app_config.urls_module:
        plugin_patterns.append(app_config.urls_module)

sorted_by_depth = sorted(all_viewsets, key=lambda vs: vs._get_nest_depth())
vs_tree = ViewSetNode()
for viewset in sorted_by_depth:
    vs_tree.add_decendent(ViewSetNode(viewset))

#: The Pulp Platform v3 API router, which can be used to manually register ViewSets with the API.
root_router = routers.DefaultRouter()

urlpatterns = [
    path(f"{API_ROOT}repair/", RepairView.as_view()),
    path(f"{API_ROOT}status/", StatusView.as_view()),
    path(
        f"{API_ROOT}orphans/cleanup/",
        OrphansCleanupViewset.as_view({"post": "cleanup"}),
    ),
    path(f"{API_ROOT}orphans/", OrphansView.as_view()),
    path(
        f"{API_ROOT}repository_versions/",
        ListRepositoryVersionViewSet.as_view({"get": "list"}),
    ),
    path(
        f"{API_ROOT}repositories/reclaim_space/",
        ReclaimSpaceViewSet.as_view({"post": "reclaim"}),
    ),
    path(
        f"{API_ROOT}importers/core/pulp/import-check/",
        PulpImporterImportCheckView.as_view(),
    ),
    path("auth/", include("rest_framework.urls")),
    path(settings.ADMIN_SITE_URL, admin.site.urls),
]

urlpatterns.append(
    path(
        f"{API_ROOT}docs/api.json",
        SpectacularJSONAPIView.as_view(authentication_classes=[], permission_classes=[]),
        name="schema",
    )
)

urlpatterns.append(
    path(
        f"{API_ROOT}docs/api.yaml",
        SpectacularYAMLAPIView.as_view(authentication_classes=[], permission_classes=[]),
        name="schema-yaml",
    )
)

urlpatterns.append(
    path(
        f"{API_ROOT}docs/",
        SpectacularRedocView.as_view(
            authentication_classes=[],
            permission_classes=[],
            url="/pulp/api/v3/docs/api.json?include_html=1",
        ),
        name="schema-redoc",
    )
)

all_routers = [root_router] + vs_tree.register_with(root_router)
for router in all_routers:
    urlpatterns.append(path(API_ROOT, include(router.urls)))

# If plugins define a urls.py, include them into the root namespace.
for plugin_pattern in plugin_patterns:
    urlpatterns.append(path("", include(plugin_pattern)))



# Generated by Django 2.2.6 on 2020-03-17 19:43

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0022_rename_last_version'),
    ]

    operations = [
        migrations.CreateModel(
            name='Export',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('params', django.contrib.postgres.fields.jsonb.JSONField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ExportedResource',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('object_id', models.UUIDField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('export', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='exported_resources', to='core.Export')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Exporter',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(default=None)),
                ('name', models.TextField(db_index=True, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.DeleteModel(
            name='FileSystemExporter',
        ),
        migrations.AddField(
            model_name='export',
            name='exporter',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Exporter'),
        ),
        migrations.AddField(
            model_name='export',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.Task'),
        ),
        migrations.CreateModel(
            name='PulpExporter',
            fields=[
                ('exporter_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_pulp_exporter', serialize=False, to='core.Exporter')),
                ('path', models.TextField()),
                ('last_export', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='core_pulp_exporter', to='core.Export')),
                ('repositories', models.ManyToManyField(related_name='core_pulp_exporter', to='core.Repository')),
            ],
            options={
                'default_related_name': '%(app_label)s_pulp_exporter',
            },
            bases=('core.exporter',),
        ),
    ]



# https://docs.djangoproject.com/en/dev/topics/db/models/#organizing-models-in-a-package

# Must be imported first as other models depend on it
from .base import (  # noqa
    BaseModel,
    Label,
    MasterModel,
)

from .access_policy import (  # noqa
    AccessPolicy,
    AutoAddObjPermsMixin,
    AutoDeleteObjPermsMixin,
    Group,
)
from .acs import AlternateContentSource, AlternateContentSourcePath  # noqa
from .content import (  # noqa
    Artifact,
    AsciiArmoredDetachedSigningService,
    Content,
    ContentManager,
    ContentArtifact,
    PulpTemporaryFile,
    RemoteArtifact,
    SigningService,
    UnsupportedDigestValidationError,
)
from .generic import GenericRelationModel  # noqa
from .exporter import (  # noqa
    Export,
    ExportedResource,
    Exporter,
    FilesystemExport,
    FilesystemExporter,
    PulpExport,
    PulpExporter,
)
from .importer import (  # noqa
    Import,
    Importer,
    PulpImport,
    PulpImporter,
)
from .publication import (  # noqa
    ContentGuard,
    Distribution,
    Publication,
    PublishedArtifact,
    PublishedMetadata,
    RBACContentGuard,
)
from .repository import (  # noqa
    Remote,
    Repository,
    RepositoryContent,
    RepositoryVersion,
    RepositoryVersionContentDetails,
)

# This can lead to circular imports with a custom user model depending on this very module
# from .role import (  # noqa
#     GroupRole,
#     Role,
#     UserRole,
# )

from .status import ContentAppStatus  # noqa

from .task import (  # noqa
    CreatedResource,
    Task,
    TaskGroup,
    Worker,
)
from .upload import (  # noqa
    Upload,
    UploadChunk,
)

# Moved here to avoid a circular import with Task
from .progress import GroupProgressReport, ProgressReport  # noqa



"""Tests related to pagination."""
import unittest
from random import randint, sample

from pulp_smash import api, config
from pulp_smash.utils import ensure_teardownclass
from pulp_smash.pulp3.utils import gen_repo, get_versions, modify_repo

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_PATH,
    FILE_MANY_FIXTURE_COUNT,
    FILE_MANY_FIXTURE_MANIFEST_URL,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import populate_pulp
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa


class RepoVersionPaginationTestCase(unittest.TestCase):
    """Test pagination of the core RepositoryVersion endpoints.

    This test case assumes that Pulp returns 100 elements in each page of
    results. This is configurable, but the current default set by all known
    Pulp installers.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.page_handler)

    def test_file_content(self):
        """Test pagination for repository versions."""
        # Add content to Pulp, create a repo, and add content to repo. We
        # sample 21 contents, because with page_size set to 10, this produces 3
        # pages, where the three three pages have unique combinations of values
        # for the "previous" and "next" links.
        populate_pulp(self.cfg, url=FILE_MANY_FIXTURE_MANIFEST_URL)
        sample_size = min(FILE_MANY_FIXTURE_COUNT, 21)
        contents = sample(self.client.get(FILE_CONTENT_PATH), sample_size)
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        for content in contents:
            modify_repo(self.cfg, repo, add_units=[content])

        # Verify pagination works for getting repo versions.
        repo = self.client.get(repo["pulp_href"])
        repo_versions = get_versions(repo, {"page_size": 10})
        self.assertEqual(len(repo_versions), sample_size + 1, repo_versions)


class PaginationTestCase(unittest.TestCase):
    """Test pagination.

    This test case assumes that Pulp returns 100 elements in each page of
    results. This is configurable, but the current default set by all known
    Pulp installers.

    This test targets the following issues:

    * `Pulp #4221 <https://pulp.plan.io/issues/4221>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.teardown_cleanups = []
        cls.number_to_create = 21

        # Perform a sanity check.
        repos = cls.client.using_handler(api.page_handler).get(FILE_REPO_PATH)
        assert len(repos) == 0, repos  # AssertEqual not available here yet

        with ensure_teardownclass(cls):
            # Create repos
            for _ in range(cls.number_to_create):
                repo = cls.client.post(FILE_REPO_PATH, gen_repo())
                cls.teardown_cleanups.append((cls.client.delete, repo["pulp_href"]))

    @classmethod
    def tearDownClass(cls):
        """Clean registered cleanups."""
        for function, argument in cls.teardown_cleanups:
            function(argument)

    def test_raw_pagination(self):
        """Assert content can be paginated page by page.

        Do the following:

        1. Without using page_handler request content
        2. Save collected_results and assert it is equal the per_page param
        3. Assert there is a next link but not a previous link
        4. Loop pages "number_to_create / per_page" (3)
        5. For each page request next link and assert length equals per_page
        6. For each page assert the presence of next and previous links
        7. Assert last page is reached
        8. Assert the final count equals number_to_create
        """

        per_page = 7  # will result in 3 pages
        resp = self.client.get(FILE_REPO_PATH, params={"limit": per_page})
        collected_results = resp["results"]
        # First call returns 7 results
        self.assertEqual(len(collected_results), per_page, collected_results)
        # no previous but there is a next
        self.assertIsNone(resp["previous"], resp["previous"])
        self.assertIsNotNone(resp["next"], resp["next"])

        # paginate pages 2 and 3
        for page in range(int(self.number_to_create / per_page)):  # [0, 1, 2]
            if page == 1:
                # there is a previous and a next
                self.assertIsNotNone(resp["previous"], resp["previous"])
                self.assertIsNotNone(resp["next"], resp["next"])
                # must have twice the size
                self.assertEqual(len(collected_results), per_page * 2, collected_results)
            if page == 2:
                # last page there is no next but there is a previous
                self.assertIsNone(resp["next"], resp["next"])
                self.assertIsNotNone(resp["previous"], resp["previous"])
                # must have 3 x the size
                self.assertEqual(len(collected_results), per_page * 3, collected_results)
                break  # last page reached
            resp = self.client.get(resp["next"])
            page_results = resp["results"]
            self.assertEqual(len(page_results), per_page, page_results)
            collected_results.extend(page_results)

        # Assert the final count
        self.assertEqual(len(collected_results), self.number_to_create, collected_results)

    @unittest.SkipTest
    def test_skip_pages(self):
        """Test jump from page 1 to page 3 (skipping page 2).

        Do the following:

        1. Request all the existing repos.
        2. Request the 7 repos on page 1 and assert equals first 7 of all.
        3. Skip page 2.
        4. Request the 7 items on page 3 and assert equals last 7 of all.
        """
        per_page = 7  # will result in 3 pages
        all_repos = self.client.using_handler(api.page_handler).get(FILE_REPO_PATH)
        self.assertEqual(len(all_repos), self.number_to_create, all_repos)

        first_page = self.client.get(FILE_REPO_PATH, params={"limit": per_page})
        last_page = self.client.get(
            FILE_REPO_PATH, params={"limit": per_page, "offset": 2 * per_page}
        )

        for index in range(0, 7):
            # Assert page 1 contains the first 7 items from _all
            self.assertEqual(first_page["results"][index], all_repos[index])

            # skip page 2

            # Assert page 3 contains the last 7 items from _all
            self.assertEqual(last_page["results"][index], all_repos[index + per_page * 2])

    def test_page_handler_pagination(self):
        """Assert page handler returns all items independent of page_size.

        This test asserts that pulp-smash page_handler will collect results
        from all pages and return it in the same call independent of the
        page_size provided.
        """
        # assert results
        repos = self.client.using_handler(api.page_handler).get(
            FILE_REPO_PATH, params={"page_size": randint(2, 11)}
        )
        self.assertEqual(len(repos), self.number_to_create, repos)



from gettext import gettext as _

from collections import defaultdict

from django.conf import settings
from django.core.exceptions import BadRequest
from django.db.models import Q, Exists, OuterRef, CharField
from django.db.models.functions import Cast
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from guardian.shortcuts import (
    get_objects_for_user as get_objects_for_user_guardian,
    get_objects_for_group as get_objects_for_group_guardian,
    get_users_with_perms as get_users_with_perms_guardian,
    get_groups_with_perms as get_groups_with_perms_guardian,
)

from pulpcore.app.models import Group
from pulpcore.app.models.role import GroupRole, Role, UserRole

User = get_user_model()


def assign_role(rolename, entity, obj=None):
    """
    Assign a role to a user or a group.

    Args:
        rolename (str): Name of the role to assign.
        entity (django.contrib.auth.User or pulpcore.app.models.Group): Entity to gain the role.
        obj (Optional[pulpcore.app.models.BaseModel]): Object the role permisssions are to be
            asserted on.
    """
    try:
        role = Role.objects.get(name=rolename)
    except Role.DoesNotExist:
        raise BadRequest(_("The role '{}' does not exist.").format(rolename))
    if obj is not None:
        ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
        if not role.permissions.filter(content_type__pk=ctype.id).exists():
            raise BadRequest(
                _("The role '{}' does not carry any permission for that object.").format(rolename)
            )
    if isinstance(entity, Group):
        GroupRole.objects.create(role=role, group=entity, content_object=obj)
    else:
        UserRole.objects.create(role=role, user=entity, content_object=obj)


def remove_role(rolename, entity, obj=None):
    """
    Remove a role from a user or a group.

    Args:
        rolename (str): Name of the role to assign.
        entity (django.contrib.auth.User or pulpcore.app.models.Group): Entity to lose the role.
        obj (Optional[pulpcore.app.models.BaseModel]): Object the role permisssions are to be
            asserted on.
    """
    try:
        role = Role.objects.get(name=rolename)
    except Role.DoesNotExist:
        raise BadRequest(_("The role '{}' does not exist.").format(rolename))
    if isinstance(entity, Group):
        qs = GroupRole.objects.filter(role=role, group=entity)
    else:
        qs = UserRole.objects.filter(role=role, user=entity)
    if obj is None:
        qs = qs.filter(object_id=None)
    else:
        ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
        qs = qs.filter(content_type__pk=ctype.id, object_id=obj.pk)
    qs.delete()


def get_perms_for_model(obj):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    return Permission.objects.filter(content_type__pk=ctype.id)


def get_objects_for_user_roles(
    user, permission_name, qs, use_groups=True, with_superuser=True, accept_global_perms=True
):
    if not user.is_authenticated:
        return qs.none()
    if with_superuser and user.is_superuser:
        return qs
    if "." in permission_name:
        app_label, codename = permission_name.split(".", maxsplit=1)
        permission = Permission.objects.get(content_type__app_label=app_label, codename=codename)
    else:
        permission = Permission.objects.get(codename=permission_name)

    if accept_global_perms:
        if user.object_roles.filter(object_id=None, role__permissions=permission).exists():
            return qs
        if (
            use_groups
            and GroupRole.objects.filter(
                group__in=user.groups.all(), object_id=None, role__permissions=permission
            ).exists()
        ):
            return qs

    user_role_pks = user.object_roles.filter(role__permissions=permission).values_list(
        "object_id", flat=True
    )
    if use_groups:
        group_role_pks = GroupRole.objects.filter(
            group__in=user.groups.all(), role__permissions=permission
        ).values_list("object_id", flat=True)

        return qs.annotate(pk_str=Cast("pk", output_field=CharField())).filter(
            Q(pk_str__in=user_role_pks) | Q(pk_str__in=group_role_pks)
        )
    else:
        return qs.annotate(pk_str=Cast("pk", output_field=CharField())).filter(
            pk_str__in=user_role_pks
        )


def get_objects_for_user(
    user, perms, qs, use_groups=True, any_perm=False, with_superuser=True, accept_global_perms=True
):
    new_qs = qs.none()
    replace = False
    if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        new_qs |= get_objects_for_user_guardian(
            user,
            perms,
            klass=qs,
            use_groups=use_groups,
            any_perm=any_perm,
            with_superuser=with_superuser,
            accept_global_perms=accept_global_perms,
        )
        replace = True
    if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        if isinstance(perms, str):
            permission_name = perms
            new_qs |= get_objects_for_user_roles(
                user,
                permission_name,
                qs=qs,
                use_groups=use_groups,
                with_superuser=with_superuser,
                accept_global_perms=accept_global_perms,
            )
        else:
            # Emulate multiple permissions and `any_perm`
            if any_perm:
                aggregate_qs = qs.none()
                for permission_name in perms:
                    aggregate_qs |= get_objects_for_user_roles(
                        user,
                        permission_name,
                        qs=qs,
                        use_groups=use_groups,
                        with_superuser=with_superuser,
                        accept_global_perms=accept_global_perms,
                    )
            else:
                aggregate_qs = qs.all()
                for permission_name in perms:
                    aggregate_qs &= get_objects_for_user_roles(
                        user,
                        permission_name,
                        qs=qs,
                        use_groups=use_groups,
                        with_superuser=with_superuser,
                        accept_global_perms=accept_global_perms,
                    )
            new_qs |= aggregate_qs
        replace = True
    if replace:
        qs = new_qs
    return qs


def get_objects_for_group_roles(group, permission_name, qs, accept_global_perms=True):
    if "." in permission_name:
        app_label, codename = permission_name.split(".", maxsplit=1)
        permission = Permission.objects.get(content_type__app_label=app_label, codename=codename)
    else:
        permission = Permission.objects.get(codename=permission_name)

    if (
        accept_global_perms
        and group.object_roles.filter(object_id=None, role__permissions=permission).exists()
    ):
        return qs

    group_role_pks = group.object_roles.filter(role__permissions=permission).values_list(
        "object_id", flat=True
    )

    return qs.annotate(pk_str=Cast("pk", output_field=CharField())).filter(
        pk_str__in=group_role_pks
    )


def get_objects_for_group(group, perms, qs, any_perm=False, accept_global_perms=True):
    new_qs = qs.none()
    replace = False
    if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        new_qs |= get_objects_for_group_guardian(
            group, perms, klass=qs, any_perm=any_perm, accept_global_perms=accept_global_perms
        )
        replace = True
    if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        if isinstance(perms, str):
            permission_name = perms
            new_qs |= get_objects_for_group_roles(
                group, permission_name, qs=qs, accept_global_perms=accept_global_perms
            )
        else:
            # Emulate multiple permissions and `any_perm`
            if any_perm:
                aggregate_qs = qs.none()
                for permission_name in perms:
                    aggregate_qs |= get_objects_for_group_roles(
                        group,
                        permission_name,
                        qs=qs,
                        accept_global_perms=accept_global_perms,
                    )
            else:
                aggregate_qs = qs.all()
                for permission_name in perms:
                    aggregate_qs &= get_objects_for_group_roles(
                        group,
                        permission_name,
                        qs=qs,
                        accept_global_perms=accept_global_perms,
                    )
            new_qs |= aggregate_qs
        replace = True
    if replace:
        qs = new_qs
    return qs


def get_users_with_perms_roles(
    obj, with_superusers=False, with_group_users=True, only_with_perms_in=None
):
    qs = User.objects.none()
    if with_superusers:
        qs |= User.objects.filter(is_superuser=True)
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    user_roles = UserRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    qs |= User.objects.filter(Exists(user_roles.filter(user=OuterRef("pk"))))
    if with_group_users:
        group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
            Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
        )
        groups = Group.objects.filter(Exists(group_roles.filter(group=OuterRef("pk")))).distinct()
        qs |= User.objects.filter(groups__in=groups)
    return qs.distinct()


def get_users_with_perms_attached_perms(
    obj, with_superusers=False, with_group_users=True, only_with_perms_in=None
):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    user_roles = UserRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    res = defaultdict(set)
    if with_superusers:
        for user in User.objects.filter(is_superuser=True):
            res[user].update(perms)
    for user_role in user_roles:
        res[user_role.user].update(
            user_role.role.permissions.filter(pk__in=perms).values_list("codename", flat=True)
        )
    if with_group_users:
        group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
            Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
        )
        for group_role in group_roles:
            for user in group_role.group.user_set.all():
                res[user].update(
                    group_role.role.permissions.filter(pk__in=perms).values_list(
                        "codename", flat=True
                    )
                )
    return {k: list(v) for k, v in res.items()}


def get_users_with_perms_attached_roles(obj, with_group_users=True, only_with_perms_in=None):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    user_roles = UserRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    res = defaultdict(set)
    for user_role in user_roles:
        res[user_role.user].update(user_role.role.name)
    if with_group_users:
        group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
            Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
        )
        for group_role in group_roles:
            for user in group_role.group.user_set.all():
                res[user].update(group_role.role.name)
    return {k: list(v) for k, v in res.items()}


# Interface copied from django guardian
def get_users_with_perms(
    obj, attach_perms=False, with_superusers=False, with_group_users=True, only_with_perms_in=None
):
    if attach_perms:
        res = defaultdict(set)
        if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            for key, value in get_users_with_perms_guardian(
                obj,
                attach_perms=True,
                with_superusers=with_superusers,
                with_group_users=with_group_users,
                only_with_perms_in=only_with_perms_in,
            ).items():
                res[key].update(value)
        if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            for key, value in get_users_with_perms_attached_perms(
                obj,
                with_superusers=with_superusers,
                with_group_users=with_group_users,
                only_with_perms_in=only_with_perms_in,
            ).items():
                res[key].update(value)
        return {k: list(v) for k, v in res.items()}
    qs = User.objects.none()
    if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        qs |= get_users_with_perms_guardian(
            obj,
            attach_perms=False,
            with_superusers=with_superusers,
            with_group_users=with_group_users,
            only_with_perms_in=only_with_perms_in,
        )
    if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
        qs |= get_users_with_perms_roles(
            obj,
            with_superusers=with_superusers,
            with_group_users=with_group_users,
            only_with_perms_in=only_with_perms_in,
        )
    return qs.distinct()


def get_groups_with_perms_roles(obj, only_with_perms_in=None):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    qs = Group.objects.filter(Exists(group_roles.filter(group=OuterRef("pk")))).distinct()
    return qs


def get_groups_with_perms_attached_perms(obj, only_with_perms_in=None):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    res = defaultdict(set)
    for group_role in group_roles:
        res[group_role.group].update(
            group_role.role.permissions.filter(pk__in=perms).values_list("codename", flat=True)
        )
    return {k: list(v) for k, v in res.items()}


def get_groups_with_perms_attached_roles(obj, only_with_perms_in=None):
    ctype = ContentType.objects.get_for_model(obj, for_concrete_model=False)
    perms = Permission.objects.filter(content_type__pk=ctype.id)
    if only_with_perms_in:
        codenames = [
            split_perm[-1]
            for split_perm in (perm.split(".", maxsplit=1) for perm in only_with_perms_in)
            if len(split_perm) == 1 or split_perm[0] == ctype.app_label
        ]
        perms = perms.filter(codename__in=codenames)
    group_roles = GroupRole.objects.filter(role__permissions__in=perms).filter(
        Q(object_id=None) | Q(content_type=ctype, object_id=obj.pk)
    )
    res = defaultdict(set)
    for group_role in group_roles:
        res[group_role.group].add(group_role.role.name)
    return {k: list(v) for k, v in res.items()}


# Interface copied from django guardian
def get_groups_with_perms(obj, attach_perms=False):
    if attach_perms:
        res = defaultdict(set)
        if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            for key, value in get_groups_with_perms_guardian(obj, attach_perms=True).items():
                res[key].update(value)
        if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            for key, value in get_groups_with_perms_attached_perms(obj).items():
                res[key].update(value)
        return {k: list(v) for k, v in res.items()}
    else:
        qs = Group.objects.none()
        if "guardian.backends.ObjectPermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            qs |= get_groups_with_perms_guardian(obj, attach_perms=False)
        if "pulpcore.backends.ObjectRolePermissionBackend" in settings.AUTHENTICATION_BACKENDS:
            qs |= get_groups_with_perms_roles(obj)
        return qs.distinct()



from pulpcore.app.role_util import (  # noqa
    assign_role,
    get_groups_with_perms,
    get_groups_with_perms_attached_perms,
    get_groups_with_perms_attached_roles,
    get_objects_for_group,
    get_objects_for_user,
    get_perms_for_model,
    get_users_with_perms,
    get_users_with_perms_attached_perms,
    get_users_with_perms_attached_roles,
    remove_role,
)



# Generated by Django 2.2.24 on 2021-06-21 19:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0066_download_concurrency_and_retry_changes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskreservedresource',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.ReservedResource'),
        ),
    ]



# Generated by Django 3.2.8 on 2021-10-21 10:58

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0081_reapplabel_group_permissions'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='task',
            options={'permissions': [('manage_roles_task', 'Can manage role assignments on task')]},
        ),
    ]



import gettext
import logging

from asgiref.sync import sync_to_async
from aiohttp.web import middleware
from django.db.utils import InterfaceError, OperationalError
from django.http.request import HttpRequest
from rest_framework.views import APIView
from rest_framework.exceptions import APIException

from .handler import Handler

log = logging.getLogger(__name__)
_ = gettext.gettext


@middleware
async def authenticate(request, handler):
    """Authenticates the request to the content app using the DRF authentication classes"""
    django_request = convert_request(request)
    fake_view = APIView()

    def _authenticate_blocking():
        drf_request = fake_view.initialize_request(django_request)
        try:
            try:
                fake_view.perform_authentication(drf_request)
            except (InterfaceError, OperationalError):
                Handler._reset_db_connection()
                fake_view.perform_authentication(drf_request)
        except APIException as e:
            log.warning(_('"{} {}" "{}": {}').format(request.method, request.path, request.host, e))

        return drf_request

    auth_request = await sync_to_async(_authenticate_blocking)()
    request["user"] = auth_request.user
    request["auth"] = auth_request.auth
    request["drf_request"] = auth_request

    return await handler(request)


def convert_request(request):
    """
    Converts an aiohttp Request to a Django HttpRequest

    This does not convert the async body to a sync body for POST requests
    """
    djr = HttpRequest()
    djr.method = request.method
    upper_keys = {k.upper() for k in request.headers.keys()}
    # These two headers are specially set by Django without the HTTP prefix
    h = {"CONTENT_LENGTH", "CONTENT_TYPE"}
    djr.META = {f"HTTP_{k}" if k not in h else k: request.headers[k] for k in upper_keys}
    djr.COOKIES = request.cookies
    djr.path = request.path
    djr.path_info = request.match_info.get("path", "")
    djr.encoding = request.charset
    djr.GET = request.query

    djr._get_scheme = lambda: request.scheme
    return djr



# Generated by Django 2.2.17 on 2020-11-18 09:41

from django.db import migrations


def namespace_access_policies_up(apps, schema_editor):
    AccessPolicy = apps.get_model('core', 'AccessPolicy')
    task_policy = AccessPolicy.objects.get(viewset_name="TaskViewSet")
    task_policy.viewset_name = "tasks"
    task_policy.save()


def namespace_access_policies_down(apps, schema_editor):
    AccessPolicy = apps.get_model('core', 'AccessPolicy')
    task_policy = AccessPolicy.objects.get(viewset_name="tasks").delete()
    task_policy.viewset_name = "TaskViewSet"
    task_policy.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0049_add_file_field_to_uploadchunk'),
    ]

    operations = [
        migrations.RunPython(namespace_access_policies_up, reverse_code=namespace_access_policies_down, elidable=True),
    ]



import os

from urllib.parse import urlparse

import aiofiles

from .base import BaseDownloader, DownloadResult


class FileDownloader(BaseDownloader):
    """
    A downloader for downloading files from the filesystem.

    It provides digest and size validation along with computation of the digests needed to save the
    file as an Artifact. It writes a new file to the disk and the return path is included in the
    :class:`~pulpcore.plugin.download.DownloadResult`.

    This downloader has all of the attributes of
    :class:`~pulpcore.plugin.download.BaseDownloader`
    """

    def __init__(self, url, *args, **kwargs):
        """
        Download files from a url that starts with `file://`

        Args:
            url (str): The url to the file. This is expected to begin with `file://`
            kwargs (dict): This accepts the parameters of
                :class:`~pulpcore.plugin.download.BaseDownloader`.

        Raises:
            ValidationError: When the url starts with `file://`, but is not a subfolder of a path in
                the ALLOWED_IMPORT_PATH setting.
        """
        from pulpcore.app.serializers import RemoteSerializer

        RemoteSerializer().validate_url(url)
        p = urlparse(url)
        self._path = os.path.abspath(os.path.join(p.netloc, p.path))
        super().__init__(url, *args, **kwargs)

    async def _run(self, extra_data=None):
        """
        Read, validate, and compute digests on the `url`. This is a coroutine.

        This method provides the same return object type and documented in
        :meth:`~pulpcore.plugin.download.BaseDownloader._run`.

        Args:
            extra_data (dict): Extra data passed to the downloader.
        """
        async with aiofiles.open(self._path, "rb") as f_handle:
            while True:
                chunk = await f_handle.read(1048576)  # 1 megabyte
                if not chunk:
                    await self.finalize()
                    break  # the reading is done
                await self.handle_data(chunk)
            return DownloadResult(
                path=self._path,
                artifact_attributes=self.artifact_attributes,
                url=self.url,
                headers=None,
            )



# Generated by Django 2.2.13 on 2020-07-14 20:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0037_pulptemporaryfile'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='remote',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote'),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-11 17:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0010_pulp_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentartifact',
            name='relative_path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='relative_path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='relative_path',
            field=models.TextField(),
        ),
    ]



# Generated by Django 2.2.20 on 2021-04-14 14:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0062_add_new_distribution_mastermodel'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='retained_versions',
            field=models.PositiveIntegerField(default=None, null=True),
        ),
    ]



"""Tests that CRUD repositories."""
import json
import re
import time
import unittest
from itertools import permutations
from urllib.parse import urljoin

from pulp_smash import api, cli, config, utils
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_repo

from requests.exceptions import HTTPError

from pulpcore.tests.functional.api.using_plugin.utils import gen_file_remote
from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_FIXTURE_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from pulpcore.tests.functional.utils import skip_if

from pulpcore.client.pulp_file.exceptions import ApiException
from pulpcore.client.pulp_file import (
    ApiClient as FileApiClient,
    FileFileRemote,
    RemotesFileApi,
)


class CRUDRepoTestCase(unittest.TestCase):
    """CRUD repositories."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.repo = {}

    def setUp(self):
        """Create an API client."""
        self.client = api.Client(self.cfg, api.json_handler)

    def test_01_create_repo(self):
        """Create repository."""
        type(self).repo = self.client.post(FILE_REPO_PATH, gen_repo())

    @skip_if(bool, "repo", False)
    def test_02_create_same_name(self):
        """Try to create a second repository with an identical name.

        * `Pulp Smash #882 <https://github.com/pulp/pulp-smash/issues/882>`_.
        * `Pulp Smash #1055
        <https://github.com/pulp/pulp-smash/issues/1055>`_.
        """
        self.client.response_handler = api.echo_handler
        response = self.client.post(FILE_REPO_PATH, gen_repo(name=self.repo["name"]))
        self.assertIn("unique", response.json()["name"][0])
        self.assertEqual(response.status_code, 400)

    @skip_if(bool, "repo", False)
    def test_02_read_repo(self):
        """Read a repository by its href."""
        repo = self.client.get(self.repo["pulp_href"])
        for key, val in self.repo.items():
            with self.subTest(key=key):
                self.assertEqual(repo[key], val)

    @skip_if(bool, "repo", False)
    def test_02_read_repo_with_specific_fields(self):
        """Read a repository by its href providing specific field list.

        Permutate field list to ensure different combinations on result.
        """
        fields = (
            "pulp_href",
            "pulp_created",
            "versions_href",
            "latest_version_href",
            "name",
            "description",
        )
        for field_pair in permutations(fields, 2):
            # ex: field_pair = ('pulp_href', 'created')
            with self.subTest(field_pair=field_pair):
                repo = self.client.get(
                    self.repo["pulp_href"], params={"fields": ",".join(field_pair)}
                )
                self.assertEqual(sorted(field_pair), sorted(repo.keys()))

    @skip_if(bool, "repo", False)
    def test_02_read_repo_without_specific_fields(self):
        """Read a repo by its href excluding specific fields."""
        # requests doesn't allow the use of != in parameters.
        url = "{}?exclude_fields=created,name".format(self.repo["pulp_href"])
        repo = self.client.get(url)
        response_fields = repo.keys()
        self.assertNotIn("created", response_fields)
        self.assertNotIn("name", response_fields)

    @skip_if(bool, "repo", False)
    def test_02_read_repos(self):
        """Read the repository by its name."""
        page = self.client.get(FILE_REPO_PATH, params={"name": self.repo["name"]})
        self.assertEqual(len(page["results"]), 1)
        for key, val in self.repo.items():
            with self.subTest(key=key):
                self.assertEqual(page["results"][0][key], val)

    @skip_if(bool, "repo", False)
    def test_02_read_all_repos(self):
        """Ensure name is displayed when listing repositories.

        See Pulp #2824 <https://pulp.plan.io/issues/2824>`_
        """
        for repo in self.client.get(FILE_REPO_PATH)["results"]:
            self.assertIsNotNone(repo["name"])

    @skip_if(bool, "repo", False)
    def test_03_fully_update_name(self):
        """Update a repository's name using HTTP PUT.

        See: `Pulp #3101 <https://pulp.plan.io/issues/3101>`_
        """
        self.do_fully_update_attr("name")

    @skip_if(bool, "repo", False)
    def test_03_fully_update_desc(self):
        """Update a repository's description using HTTP PUT."""
        self.do_fully_update_attr("description")

    def do_fully_update_attr(self, attr):
        """Update a repository attribute using HTTP PUT.

        :param attr: The name of the attribute to update. For example,
            "description." The attribute to update must be a string.
        """
        repo = self.client.get(self.repo["pulp_href"])
        string = utils.uuid4()
        repo[attr] = string
        self.client.put(repo["pulp_href"], repo)

        # verify the update
        repo = self.client.get(repo["pulp_href"])
        self.assertEqual(string, repo[attr])

    @skip_if(bool, "repo", False)
    def test_03_partially_update_name(self):
        """Update a repository's name using HTTP PATCH.

        See: `Pulp #3101 <https://pulp.plan.io/issues/3101>`_
        """
        self.do_partially_update_attr("name")

    @skip_if(bool, "repo", False)
    def test_03_partially_update_desc(self):
        """Update a repository's description using HTTP PATCH."""
        self.do_partially_update_attr("description")

    def do_partially_update_attr(self, attr):
        """Update a repository attribute using HTTP PATCH.

        :param attr: The name of the attribute to update. For example,
            "description." The attribute to update must be a string.
        """
        string = utils.uuid4()
        self.client.patch(self.repo["pulp_href"], {attr: string})

        # verify the update
        repo = self.client.get(self.repo["pulp_href"])
        self.assertEqual(repo[attr], string)

    @skip_if(bool, "repo", False)
    def test_03_set_remote_on_repository(self):
        """Test setting remotes on repositories."""
        body = gen_file_remote()
        remote = self.client.post(FILE_REMOTE_PATH, body)

        # verify that syncing with no remote raises an error
        with self.assertRaises(HTTPError):
            self.client.post(urljoin(self.repo["pulp_href"], "sync/"))

        # test setting the remote on the repo
        self.client.patch(self.repo["pulp_href"], {"remote": remote["pulp_href"]})

        # test syncing without a remote
        self.client.post(urljoin(self.repo["pulp_href"], "sync/"))

        repo = self.client.get(self.repo["pulp_href"])
        self.assertEqual(repo["latest_version_href"], f"{repo['pulp_href']}versions/1/")

    @skip_if(bool, "repo", False)
    def test_04_delete_repo(self):
        """Delete a repository."""
        self.client.delete(self.repo["pulp_href"])

        # verify the delete
        with self.assertRaises(HTTPError):
            self.client.get(self.repo["pulp_href"])

    def test_negative_create_repo_with_invalid_parameter(self):
        """Attempt to create repository passing extraneous invalid parameter.

        Assert response returns an error 400 including ["Unexpected field"].
        """
        response = api.Client(self.cfg, api.echo_handler).post(FILE_REPO_PATH, gen_repo(foo="bar"))
        assert response.status_code == 400
        assert response.json()["foo"] == ["Unexpected field"]


class CRUDRemoteTestCase(unittest.TestCase):
    """CRUD remotes."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()

    def setUp(self):
        self.client = FileApiClient(self.cfg.get_bindings_config())
        self.remotes_api = RemotesFileApi(self.client)
        self.remote_attrs = {
            "name": utils.uuid4(),
            "url": FILE_FIXTURE_MANIFEST_URL,
            "ca_cert": None,
            "client_cert": None,
            "client_key": None,
            "tls_validation": False,
            "proxy_url": None,
            "username": "pulp",
            "password": "pulp",
            "download_concurrency": 10,
            "policy": "on_demand",
            "total_timeout": None,
            "connect_timeout": None,
            "sock_connect_timeout": None,
            "sock_read_timeout": None,
        }
        self.remote = self.remotes_api.create(self.remote_attrs)

    def _compare_results(self, data, received):
        self.assertFalse(hasattr(received, "password"))

        # handle write only fields
        data.pop("username", None)
        data.pop("password", None)
        data.pop("client_key", None)

        for k in data:
            self.assertEqual(getattr(received, k), data[k])

    def test_read(self):
        # Compare initial-attrs vs remote created in setUp
        self._compare_results(self.remote_attrs, self.remote)

    def test_update(self):
        data = {"download_concurrency": 23, "policy": "immediate"}
        self.remotes_api.partial_update(self.remote.pulp_href, data)
        time.sleep(1)  # without this, the read returns the pre-patch values
        new_remote = self.remotes_api.read(self.remote.pulp_href)
        self._compare_results(data, new_remote)

    def test_password_writeable(self):
        """Test that a password can be updated with a PUT request."""
        cli_client = cli.Client(self.cfg)
        remote = self.remotes_api.create({"name": "test_pass", "url": "http://", "password": "new"})
        href = remote.pulp_href
        uuid = re.search(r"/pulp/api/v3/remotes/file/file/([\w-]+)/", href).group(1)
        shell_cmd = (
            f"import pulpcore; print(pulpcore.app.models.Remote.objects.get(pk='{uuid}').password)"
        )

        self.addCleanup(self.remotes_api.delete, href)

        # test a PUT request with a new password
        remote_update = FileFileRemote(name="test_pass", url="http://", password="changed")
        response = self.remotes_api.update(href, remote_update)
        monitor_task(response.task)
        exc = cli_client.run(["pulpcore-manager", "shell", "-c", shell_cmd])
        self.assertEqual(exc.stdout.rstrip("\n"), "changed")

    def test_password_not_unset(self):
        """Test that password doesn't get unset when not passed with a PUT request."""
        cli_client = cli.Client(self.cfg)
        remote = self.remotes_api.create({"name": "test_pass", "url": "http://", "password": "new"})
        href = remote.pulp_href
        uuid = re.search(r"/pulp/api/v3/remotes/file/file/([\w-]+)/", href).group(1)
        shell_cmd = (
            f"import pulpcore; print(pulpcore.app.models.Remote.objects.get(pk='{uuid}').password)"
        )

        self.addCleanup(self.remotes_api.delete, href)

        # test a PUT request without a password
        remote_update = FileFileRemote(name="pass_test", url="http://")
        response = self.remotes_api.update(href, remote_update)
        monitor_task(response.task)
        exc = cli_client.run(["pulpcore-manager", "shell", "-c", shell_cmd])
        self.assertEqual(exc.stdout.rstrip("\n"), "new")

    def test_timeout_attributes(self):
        # Test valid timeout settings (float >= 0)
        data = {
            "total_timeout": 1.0,
            "connect_timeout": 66.0,
            "sock_connect_timeout": 0.0,
            "sock_read_timeout": 3.1415926535,
        }
        self.remotes_api.partial_update(self.remote.pulp_href, data)
        time.sleep(1)
        new_remote = self.remotes_api.read(self.remote.pulp_href)
        self._compare_results(data, new_remote)

    def test_timeout_attributes_float_lt_zero(self):
        # Test invalid float < 0
        data = {
            "total_timeout": -1.0,
        }
        with self.assertRaises(ApiException):
            self.remotes_api.partial_update(self.remote.pulp_href, data)

    def test_timeout_attributes_non_float(self):
        # Test invalid non-float
        data = {
            "connect_timeout": "abc",
        }
        with self.assertRaises(ApiException):
            self.remotes_api.partial_update(self.remote.pulp_href, data)

    def test_timeout_attributes_reset_to_empty(self):
        # Test reset to empty
        data = {
            "total_timeout": False,
            "connect_timeout": None,
            "sock_connect_timeout": False,
            "sock_read_timeout": None,
        }
        response = self.remotes_api.partial_update(self.remote.pulp_href, data)
        monitor_task(response.task)
        new_remote = self.remotes_api.read(self.remote.pulp_href)
        self._compare_results(data, new_remote)

    def test_delete(self):
        response = self.remotes_api.delete(self.remote.pulp_href)
        monitor_task(response.task)
        # verify the delete
        with self.assertRaises(ApiException):
            self.remotes_api.read(self.remote.pulp_href)

    def test_headers(self):
        # Test that headers value must be a list of dicts
        data = {"headers": {"Connection": "keep-alive"}}
        with self.assertRaises(ApiException):
            self.remotes_api.partial_update(self.remote.pulp_href, data)
        data = {"headers": [1, 2, 3]}
        with self.assertRaises(ApiException):
            self.remotes_api.partial_update(self.remote.pulp_href, data)
        data = {"headers": [{"Connection": "keep-alive"}]}
        self.remotes_api.partial_update(self.remote.pulp_href, data)


class CreatePulpLabelsRemoteTestCase(unittest.TestCase):
    """A test case for verifying whether pulp_labels are correctly assigned to a new remote."""

    @classmethod
    def setUpClass(cls):
        """Initialize class-wide variables"""
        cls.cfg = config.get_config()

        cls.api_client = api.Client(cls.cfg, api.json_handler)
        cls.file_client = FileApiClient(cls.cfg.get_bindings_config())
        cls.remotes_api = RemotesFileApi(cls.file_client)

        cls.pulp_labels = {"environment": "dev"}

    def test_create_remote(self):
        """Test if a created remote contains pulp_labels when passing JSON data."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": FILE_FIXTURE_MANIFEST_URL,
            "pulp_labels": self.pulp_labels,
        }
        remote = self.remotes_api.create(remote_attrs)
        self.addCleanup(self.remotes_api.delete, remote.pulp_href)

        self.assertEqual(remote.pulp_labels, self.pulp_labels)

    def test_create_remote_using_form(self):
        """Test if a created remote contains pulp_labels when passing form data."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": FILE_FIXTURE_MANIFEST_URL,
            "pulp_labels": json.dumps(self.pulp_labels),
        }
        remote = self.api_client.post(FILE_REMOTE_PATH, data=remote_attrs)
        self.addCleanup(self.remotes_api.delete, remote["pulp_href"])
        self.assertEqual(remote["pulp_labels"], self.pulp_labels)


class RemoteFileURLsValidationTestCase(unittest.TestCase):
    """A test case that verifies the validation of remotes' URLs."""

    @classmethod
    def setUpClass(cls):
        """Initialize class-wide variables"""
        cls.cfg = config.get_config()

        cls.api_client = api.Client(cls.cfg, api.json_handler)
        cls.file_client = FileApiClient(cls.cfg.get_bindings_config())
        cls.remotes_api = RemotesFileApi(cls.file_client)

    def test_invalid_absolute_pathname(self):
        """Test the validation of an invalid absolute pathname."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": "file://error/path/name",
        }
        self.raise_for_invalid_request(remote_attrs)

    def test_invalid_import_path(self):
        """Test the validation of an invalid import pathname."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": "file:///error/path/name",
        }
        self.raise_for_invalid_request(remote_attrs)

    def raise_for_invalid_request(self, remote_attrs):
        """Check if Pulp returns HTTP 400 after issuing an invalid request."""
        with self.assertRaises(ApiException) as ae:
            remote = self.remotes_api.create(remote_attrs)
            self.addCleanup(self.remotes_api.delete, remote.pulp_href)

        self.assertEqual(ae.exception.status, 400)

    def test_valid_import_path(self):
        """Test the creation of a remote after passing a valid URL."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": "file:///tmp/good",
        }

        remote = self.remotes_api.create(remote_attrs)
        self.addCleanup(self.remotes_api.delete, remote.pulp_href)

    def test_no_username_password(self):
        """Test that the remote url can't contain username/password."""
        remote_attrs = {
            "name": utils.uuid4(),
            "url": "http://elladan@rivendell.org",
        }
        self.raise_for_invalid_request(remote_attrs)

        remote_attrs = {
            "name": utils.uuid4(),
            "url": "http://elladan:pass@rivendell.org",
        }
        self.raise_for_invalid_request(remote_attrs)



from gettext import gettext as _

from rest_framework import fields, serializers

from django.conf import settings

from pulpcore.app.models import Content
from pulpcore.app.serializers import ValidateFieldsMixin


class OrphansCleanupSerializer(serializers.Serializer, ValidateFieldsMixin):

    content_hrefs = fields.ListField(
        required=False,
        help_text=_("Will delete specified content and associated Artifacts if they are orphans."),
    )
    orphan_protection_time = serializers.IntegerField(
        help_text=(
            "The time in minutes for how long Pulp will hold orphan Content and Artifacts before "
            "they become candidates for deletion by this orphan cleanup task. This should ideally "
            "be longer than your longest running task otherwise any content created during that "
            "task could be cleaned up before the task finishes. If not specified, default is used "
            "from settings which is 1440 minutes (24 hours)"
        ),
        allow_null=True,
        required=False,
        default=settings.ORPHAN_PROTECTION_TIME,
    )

    def validate_content_hrefs(self, value):
        """
        Check that the content_hrefs is not an empty list and contains all valid hrefs.
        Args:
            value (list): The list supplied by the user
        Returns:
            The list of pks (not hrefs) after validation
        Raises:
            ValidationError: If the list is empty or contains invalid hrefs.
        """
        if len(value) == 0:
            raise serializers.ValidationError("Must not be [].")
        from pulpcore.app.viewsets import NamedModelViewSet

        pks_to_return = []
        for href in value:
            pks_to_return.append(NamedModelViewSet.get_resource(href, Content).pk)

        return pks_to_return



import os
from gettext import gettext as _

from django.core.files.uploadedfile import TemporaryUploadedFile
from django.core.files.uploadhandler import TemporaryFileUploadHandler
from pygtrie import StringTrie

from pulpcore.app import models
from pulpcore.app import pulp_hashlib


class PulpTemporaryUploadedFile(TemporaryUploadedFile):
    """
    A file uploaded to a temporary location in Pulp.
    """

    def __init__(self, name, content_type, size, charset, content_type_extra=None):
        self.hashers = {}
        for hasher in models.Artifact.DIGEST_FIELDS:
            self.hashers[hasher] = pulp_hashlib.new(hasher)
        super().__init__(name, content_type, size, charset, content_type_extra)

    @classmethod
    def from_file(cls, file):
        """
        Create a PulpTemporaryUploadedFile from a file system file

        Args:
            file (File): a filesystem file

        Returns:
            PulpTemporaryUploadedFile: instantiated instance from file
        """
        name = os.path.basename(file.name)
        instance = cls(name, "", file.size, "", "")
        instance.file = file
        data = file.read()

        # calling the method read() moves the file's pointer to the end of the file object,
        # thus, it is necessary to reset the file's pointer position back to 0 in case of
        # calling the method read() again from another place
        file.seek(0)

        for hasher in models.Artifact.DIGEST_FIELDS:
            instance.hashers[hasher].update(data)
        return instance


class HashingFileUploadHandler(TemporaryFileUploadHandler):
    """
    Upload handler that streams data into a temporary file.
    """

    def new_file(
        self,
        field_name,
        file_name,
        content_type,
        content_length,
        charset=None,
        content_type_extra=None,
    ):
        """
        Signal that a new file has been started.

        Args:
            field_name (str): Name of the model field that this file is associated with. This
                value is not used by this implementation of TemporaryFileUploadHandler.
            file_name (str): Name of file being uploaded.
            content_type (str): Type of file
            content_length (int): Size of the file being stored. This value is not used by this
                implementation of TemporaryFileUploadHandler.
            charset (str):
        """
        self.field_name = field_name
        self.content_length = content_length
        self.file = PulpTemporaryUploadedFile(
            file_name, content_type, 0, charset, content_type_extra
        )

    def receive_data_chunk(self, raw_data, start):
        self.file.write(raw_data)
        for hasher in models.Artifact.DIGEST_FIELDS:
            self.file.hashers[hasher].update(raw_data)


class TemporaryDownloadedFile(TemporaryUploadedFile):
    """
    A temporary downloaded file.

    The FileSystemStorage backend treats this object the same as a TemporaryUploadedFile. The
    storage backend attempts to link the file to its final location. If the final location is on a
    different physical drive, the file is copied to its final destination.
    """

    def __init__(self, file, name=None):
        """
        A constructor that does not create a blank temporary file.

        The __init__ for TemporaryUploadedFile creates an empty temporary file. This constructor
        is designed to handle files that have already been written to disk.

        Args:
            file (file): An open file
            name (str): Name of the file
        """
        self.file = file
        if name is None:
            name = getattr(file, "name", None)
        self.name = name


def validate_file_paths(paths):
    """
    Check for valid POSIX paths (ie ones that aren't duplicated and don't overlap).

    Overlapping paths are where one path terminates inside another (e.g. a/b and a/b/c).

    This function will raise an exception at the first dupe or overlap it detects. We use a trie (or
    prefix tree) to keep track of which paths we've already seen.

    Args:
        paths (iterable of str): An iterable of strings each representing a relative path

    Raises:
        ValueError: If any path overlaps another
    """
    overlap_error = _("The path for file '{path}' overlaps: {conflicts}")

    path_trie = StringTrie(separator="/")
    dups = []
    overlaps = []
    for path in paths:
        if path in path_trie:
            # path duplicates a path already in the trie
            dups.append(path)
        elif path_trie.has_subtrie(path):
            # overlap where path is 'a/b' and trie has 'a/b/c'
            conflicts = [item[0] for item in path_trie.items(prefix=path)]
            overlaps.append(overlap_error.format(path=path, conflicts=", ".join(conflicts)))
        else:
            prefixes = list(path_trie.prefixes(path))
            if prefixes:
                # overlap where path is 'a/b/c' and trie has 'a/b'
                conflicts = [prefix.key for prefix in prefixes]
                overlaps.append(overlap_error.format(path=path, conflicts=", ".join(conflicts)))

        # if there are no overlaps, add it to our trie and continue
        path_trie[path] = True

    if dups or overlaps:
        dups_msg = ""
        overlaps_msg = ""
        if dups:
            dups_msg = _("Paths are duplicated: {paths}").format(paths=",".join(dups))
        if overlaps:
            overlaps_msg = "\n".join(overlaps)

        raise ValueError(
            _("Path errors found. {dups}\n{overlaps}").format(dups=dups_msg, overlaps=overlaps_msg)
        )



from gettext import gettext as _

from pulpcore.app.models import ContentArtifact
from pulpcore.app.files import validate_file_paths


def validate_publication_paths(publication):
    """
    Validate artifact relative paths for dupes or overlap (e.g. a/b and a/b/c).

    Raises:
        ValueError: If two artifact relative paths are dupes or overlap
    """
    paths = list(publication.published_artifact.values_list("relative_path", flat=True))

    if publication.pass_through:
        paths += ContentArtifact.objects.filter(
            content__pk__in=publication.repository_version.content
        ).values_list("relative_path", flat=True)

    try:
        validate_file_paths(paths)
    except ValueError as e:
        raise ValueError(_("Cannot create publication. {err}.").format(err=e))



"""
Tests FilesystemExporter and FilesystemExport functionality

NOTE: assumes ALLOWED_EXPORT_PATHS setting contains "/tmp" - all tests will fail if this is not
the case.
"""
import unittest
from pulp_smash import api, cli, config
from pulp_smash.utils import uuid4
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.utils import gen_repo

from pulpcore.client.pulpcore.exceptions import ApiException

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ExportersFilesystemApi,
    ExportersFilesystemExportsApi,
)
from pulpcore.client.pulp_file import (
    ContentFilesApi,
    FileFilePublication,
    PublicationsFileApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
    RepositorySyncURL,
    RemotesFileApi,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)

NUM_REPOS = 1
NUM_EXPORTERS = 4


class BaseExporterCase(unittest.TestCase):
    """
    Base functionality for Exporter and Export test classes

    The export process isn't possible without repositories having been sync'd - arranging for
    that to happen once per-class (instead of once-per-test) is the primary purpose of this parent
    class.
    """

    @classmethod
    def _setup_repositories(cls):
        """Create and sync a number of repositories to be exported."""
        # create and remember a set of repo
        repos = []
        remotes = []
        publications = []
        for r in range(NUM_REPOS):
            repo = cls.repo_api.create(gen_repo())
            remote = cls.remote_api.create(gen_file_remote())

            repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
            sync_response = cls.repo_api.sync(repo.pulp_href, repository_sync_data)
            monitor_task(sync_response.task)

            repo = cls.repo_api.read(file_file_repository_href=repo.pulp_href)
            publish_data = FileFilePublication(repository=repo.pulp_href)
            publish_response = cls.publication_api.create(publish_data)
            created_resources = monitor_task(publish_response.task).created_resources
            publication_href = created_resources[0]
            publication = cls.publication_api.read(publication_href)

            repos.append(repo)
            remotes.append(remote)
            publications.append(publication)
        return repos, remotes, publications

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.core_client = CoreApiClient(configuration=cls.cfg.get_bindings_config())
        cls.file_client = gen_file_client()

        cls.content_api = ContentFilesApi(cls.file_client)
        cls.repo_api = RepositoriesFileApi(cls.file_client)
        cls.versions_api = RepositoriesFileVersionsApi(cls.file_client)
        cls.remote_api = RemotesFileApi(cls.file_client)
        cls.publication_api = PublicationsFileApi(cls.file_client)
        cls.exporter_api = ExportersFilesystemApi(cls.core_client)
        cls.exports_api = ExportersFilesystemExportsApi(cls.core_client)

        cls.repos, cls.remotes, cls.publications = cls._setup_repositories()

    @classmethod
    def tearDownClass(cls):
        """Clean up after ourselves."""
        for remote in cls.remotes:
            cls.remote_api.delete(remote.pulp_href)
        for repo in cls.repos:
            cls.repo_api.delete(repo.pulp_href)
        delete_orphans()

    def _delete_exporter(self, exporter):
        """
        Utility routine to delete an exporter.
        """
        cli_client = cli.Client(self.cfg)
        cmd = ("rm", "-rf", exporter.path)
        cli_client.run(cmd, sudo=True)

        result = self.exporter_api.delete(exporter.pulp_href)
        monitor_task(result.task)

    def _create_exporter(self, params={}):
        """
        Utility routine to create an exporter for the available repositories.
        """
        body = {
            "name": uuid4(),
            "path": "/tmp/{}/".format(uuid4()),
        }
        body.update(params)

        exporter = self.exporter_api.create(body)
        self.addCleanup(self._delete_exporter, exporter)
        return exporter, body


class FilesystemExporterTestCase(BaseExporterCase):
    """Test FilesystemExporter CURDL methods."""

    def test_create(self):
        """Create a FilesystemExporter."""
        exporter, body = self._create_exporter()
        self.assertEqual(body["name"], exporter.name)
        self.assertEqual(body["path"], exporter.path)

    def test_read(self):
        """Read a created FilesystemExporter."""
        exporter_created, body = self._create_exporter()
        exporter_read = self.exporter_api.read(exporter_created.pulp_href)
        self.assertEqual(exporter_created.name, exporter_read.name)
        self.assertEqual(exporter_created.path, exporter_read.path)

    def test_partial_update(self):
        """Update a FilesystemExporter's path."""
        exporter_created, body = self._create_exporter()
        body = {"path": "/tmp/{}".format(uuid4())}
        result = self.exporter_api.partial_update(exporter_created.pulp_href, body)
        monitor_task(result.task)
        exporter_read = self.exporter_api.read(exporter_created.pulp_href)
        self.assertNotEqual(exporter_created.path, exporter_read.path)
        self.assertEqual(body["path"], exporter_read.path)

    def test_list(self):
        """Show a set of created FilesystemExporters."""
        starting_exporters = self.exporter_api.list().results
        for x in range(NUM_EXPORTERS):
            self._create_exporter()
        ending_exporters = self.exporter_api.list().results
        self.assertEqual(NUM_EXPORTERS, len(ending_exporters) - len(starting_exporters))

    def test_delete(self):
        """Delete a pulpExporter."""
        exporter = self.exporter_api.create({"name": "test", "path": "/tmp/abc"})
        result = self.exporter_api.delete(exporter.pulp_href)
        monitor_task(result.task)
        with self.assertRaises(ApiException) as ae:
            self.exporter_api.read(exporter.pulp_href)
        self.assertEqual(404, ae.exception.status)

    def test_method(self):
        """Test the method field."""
        exporter, _ = self._create_exporter({"method": "symlink"})
        self.assertEqual("symlink", exporter.method)

        with self.assertRaises(ApiException) as ae:
            exporter = self._create_exporter({"method": "invalid"})
        self.assertEqual(400, ae.exception.status)


class FilesystemExportTestCase(BaseExporterCase):
    """Test FilesystemExport CRDL methods (Update is not allowed)."""

    def _gen_export(self, exporter, publication):
        """Create and read back an export for the specified FilesystemExporter."""
        body = {"publication": publication.pulp_href}
        export_response = self.exports_api.create(exporter.pulp_href, body)
        monitor_task(export_response.task)

        task = self.client.get(export_response.task)
        resources = task["created_resources"]
        self.assertEqual(1, len(resources))

        return self.exports_api.read(resources[0])

    def test_export(self):
        """Issue and evaluate a FilesystemExport (tests both Create and Read)."""
        exporter, body = self._create_exporter({"method": "write"})
        export = self._gen_export(exporter, self.publications[0])
        self.assertIsNotNone(export)

    def test_list(self):
        """Find all the FilesystemExports for a FilesystemExporter."""
        exporter, body = self._create_exporter({"method": "write"})
        for i in range(NUM_REPOS):
            self._gen_export(exporter, self.publications[i])
        exporter = self.exporter_api.read(exporter.pulp_href)
        exports = self.exports_api.list(exporter.pulp_href).results
        self.assertEqual(NUM_REPOS, len(exports))

    def test_delete(self):
        """Test deleting exports for a FilesystemExporter."""
        exporter, body = self._create_exporter({"method": "write"})
        export = self._gen_export(exporter, self.publications[0])
        self.exports_api.delete(export.pulp_href)
        with self.assertRaises(ApiException) as ae:
            self.exports_api.read(export.pulp_href)
        self.assertEqual(404, ae.exception.status)



from gettext import gettext as _

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError
from django.db import IntegrityError
from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend, filters
from drf_spectacular.utils import extend_schema
from guardian.models.models import GroupObjectPermission
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.exceptions import PermissionDenied
from rest_framework.filters import OrderingFilter
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from pulpcore.app.models import Group
from pulpcore.app.models.role import GroupRole, Role, UserRole
from pulpcore.app.viewsets import BaseFilterSet, NamedModelViewSet, RolesMixin, NAME_FILTER_OPTIONS
from pulpcore.app.serializers import (
    GroupSerializer,
    GroupUserSerializer,
    GroupRoleSerializer,
    PermissionSerializer,
    RoleSerializer,
    UserSerializer,
    UserRoleSerializer,
)

User = get_user_model()


class UserFilter(BaseFilterSet):
    """
    FilterSet for User.
    """

    class Meta:
        model = User
        fields = {
            "username": ["exact", "iexact", "in", "contains", "icontains"],
            "first_name": ["exact", "iexact", "in", "contains", "icontains"],
            "last_name": ["exact", "iexact", "in", "contains", "icontains"],
            "email": ["exact", "iexact", "in", "contains", "icontains"],
            "is_active": ["exact"],
            "is_staff": ["exact"],
        }


class UserViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for User.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "users"
    router_lookup = "user"
    filterset_class = UserFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = UserSerializer
    queryset = User.objects.all()
    ordering = ("-date_joined",)

    @extend_schema(description="List user permissions.")
    @action(detail=True, methods=["get"], serializer_class=PermissionSerializer)
    def permissions(self, request, pk):
        """
        List user permissions.
        """
        user = self.get_object()
        object_permission = PermissionSerializer(
            user.userobjectpermission_set.all(), many=True, context={"request": request}
        )
        permissions = PermissionSerializer(user.user_permissions.all(), many=True)
        return Response(object_permission.data + permissions.data)


class GroupFilter(BaseFilterSet):
    """
    FilterSet for Group.
    """

    class Meta:
        model = Group
        fields = {
            "id": ["exact", "in"],
            "name": ["exact", "iexact", "in", "contains", "icontains"],
        }


class GroupViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
    RolesMixin,
):
    """
    ViewSet for Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "groups"
    router_lookup = "group"
    filterset_class = GroupFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    ordering = ("name",)
    queryset_filtering_required_permission = "core.view_group"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:core.add_group",
            },
            {
                "action": ["retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.view_group",
            },
            {
                "action": ["update", "partial_update"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.change_group",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.delete_group",
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": ["has_model_or_obj_perms:core.manage_roles_group"],
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "core.group_owner"},
            },
        ],
    }

    LOCKED_ROLES = {
        "core.group_creator": [
            "core.add_group",
        ],
        "core.group_owner": [
            "core.view_group",
            "core.change_group",
            "core.delete_group",
            "core.manage_roles_group",
        ],
        "core.group_viewer": [
            "core.view_group",
        ],
    }


class GroupModelPermissionViewSet(NamedModelViewSet):
    """
    ViewSet for Model Permissions that belong to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "model_permissions"
    router_lookup = "model_permission"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "group__pk"}
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()
    pulp_model_alias = "ModelPermission"

    def get_model_permission(self, request):
        """Get model permission"""
        data = {}
        for key, value in request.data.items():
            if key == "permission":
                if "." in value:
                    data["content_type__app_label"], data["codename"] = value.split(".", maxsplit=1)
                else:
                    data["codename"] = value
                continue

            if key == "pulp_href":
                if "id" in data.keys():
                    continue
                data["id"] = value.strip("/").split("/")[-1]
                continue

            data[key] = value

        try:
            permission = Permission.objects.get(**data)
        except (Permission.MultipleObjectsReturned, Permission.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))

        return permission

    @extend_schema(description="Retrieve a model permission from a group.")
    def retrieve(self, request, group_pk, pk):
        instance = get_object_or_404(Permission, pk=pk)
        serializer = PermissionSerializer(instance, context={"group_pk": group_pk})
        return Response(serializer.data)

    @extend_schema(description="List group permissions.", responses={200: PermissionSerializer})
    def list(self, request, group_pk):
        """
        List group model permissions.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.permissions.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = PermissionSerializer(page, context={"group_pk": group_pk}, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = PermissionSerializer(queryset, context={"group_pk": group_pk}, many=True)
        return Response(serializer.data)

    @extend_schema(
        description="Add a model permission to a group.", responses={201: PermissionSerializer}
    )
    def create(self, request, group_pk):
        """
        Add a model permission to a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = self.get_model_permission(request)
        group.permissions.add(permission)
        group.save()
        serializer = PermissionSerializer(permission, context={"group_pk": group_pk})
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    @extend_schema(description="Remove a model permission from a group.")
    def destroy(self, request, group_pk, pk):
        """
        Remove a model permission from a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = get_object_or_404(Permission, pk=pk)
        group.permissions.remove(permission)
        group.save()
        return Response(status=status.HTTP_204_NO_CONTENT)


class GroupObjectPermissionViewSet(NamedModelViewSet):
    """
    ViewSet for Object Permissions that belong to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "object_permissions"
    router_lookup = "object_permission"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "group__pk"}
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()
    pulp_model_alias = "ObjectPermission"

    def get_object_pk(self, request):
        """Return an object's pk from the request."""

        if "obj" not in request.data:
            raise ValidationError(_("Please provide 'obj' value"))

        obj_url = request.data["obj"]
        try:
            obj = NamedModelViewSet.get_resource(obj_url)
        except ValidationError:
            raise ValidationError(_("Invalid value for 'obj': {}.").format(obj_url))

        return obj.pk

    def get_model_permission(self, request):
        """Get model permission"""

        if "permission" not in request.data:
            raise ValidationError(_("Please provide 'permission' value"))

        codename = request.data["permission"].split(".")[-1]

        try:
            permission = Permission.objects.get(codename=codename)
        except (Permission.MultipleObjectsReturned, Permission.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))

        return permission

    @extend_schema(description="Retrieve a model permission from a group.")
    def retrieve(self, request, group_pk, pk):
        instance = get_object_or_404(GroupObjectPermission, pk=pk)
        serializer = PermissionSerializer(
            instance, context={"group_pk": group_pk, "request": request}
        )
        return Response(serializer.data)

    @extend_schema(
        description="List group object permissions.", responses={200: PermissionSerializer}
    )
    def list(self, request, group_pk):
        """
        List group object permissions.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.groupobjectpermission_set.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = PermissionSerializer(
                page, context={"group_pk": group_pk, "request": request}, many=True
            )
            return self.get_paginated_response(serializer.data)

        serializer = PermissionSerializer(
            queryset, context={"group_pk": group_pk, "request": request}, many=True
        )
        return Response(serializer.data)

    @extend_schema(
        description="Add an object permission to a group.", responses={201: PermissionSerializer}
    )
    def create(self, request, group_pk):
        """
        Create an object permission to a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = self.get_model_permission(request)
        object_pk = self.get_object_pk(request)
        object_permission = GroupObjectPermission(
            group=group,
            permission=permission,
            content_type_id=permission.content_type_id,
            object_pk=object_pk,
        )

        try:
            object_permission.save()
        except IntegrityError:
            raise ValidationError(_("The assigned permission already exists."))

        serializer = PermissionSerializer(
            object_permission, context={"group_pk": group_pk, "request": request}
        )
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    @extend_schema(description="Remove an object permission from a group.")
    def destroy(self, request, group_pk, pk):
        """
        Delete an object permission from a group.
        """
        object_permission = get_object_or_404(GroupObjectPermission, pk=pk, group_id=group_pk)
        object_permission.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


class GroupUserViewSet(NamedModelViewSet):
    """
    ViewSet for Users that belong to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "users"
    router_lookup = "user"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "groups__pk"}
    serializer_class = GroupUserSerializer
    queryset = User.objects.all()

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:core.view_group",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:core.change_group",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:core.change_group",
            },
        ],
        "creation_hooks": [],
    }

    def list(self, request, group_pk):
        """
        List group users.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.user_set.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = GroupUserSerializer(page, context={"request": request}, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def create(self, request, group_pk):
        """
        Add a user to a group.
        """
        group = Group.objects.get(pk=group_pk)
        if not request.data:
            raise ValidationError(
                _("Please provide one of the following values for User: 'pk', 'id', 'username'")
            )
        try:
            user = User.objects.get(**request.data)
        except (User.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))
        group.user_set.add(user)
        group.save()
        serializer = GroupUserSerializer(user, context={"request": request})
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def destroy(self, request, group_pk, pk):
        """
        Remove a user from a group.
        """
        group = Group.objects.get(pk=group_pk)
        user = get_object_or_404(User, pk=pk)
        group.user_set.remove(user)
        group.save()
        return Response(status=status.HTTP_204_NO_CONTENT)


class RoleFilter(BaseFilterSet):
    """
    FilterSet for Role.
    """

    class Meta:
        model = Role
        fields = {
            "name": NAME_FILTER_OPTIONS,
            "locked": ["exact"],
        }


class RoleViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for Role.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "roles"
    router_lookup = "role"
    filterset_class = RoleFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = RoleSerializer
    queryset = Role.objects.all()
    ordering = ("name",)

    def get_object(self, **kwargs):
        instance = super().get_object(**kwargs)
        if instance.locked and self.request.method not in SAFE_METHODS:
            raise PermissionDenied(detail=_("The role is locked."))
        return instance


class _CharInFilter(filters.BaseInFilter, filters.CharFilter):
    pass


class NestedRoleFilter(BaseFilterSet):
    """
    FilterSet for Roles nested under users / groups.
    """

    # For some reason the generated Filters don't work here
    role = filters.CharFilter(field_name="role__name")
    role__in = _CharInFilter(field_name="role__name", lookup_expr="in")
    role__contains = filters.CharFilter(field_name="role__name", lookup_expr="contains")
    role__icontains = filters.CharFilter(field_name="role__name", lookup_expr="icontains")
    role__startswith = filters.CharFilter(field_name="role__name", lookup_expr="startswith")
    content_object = filters.CharFilter(
        label="content_object", method="content_object_filter_function"
    )

    def content_object_filter_function(self, queryset, name, value):
        if value == "null":
            return queryset.filter(object_id=None)
        else:
            try:
                obj = NamedModelViewSet.get_resource(value)
            except ValidationError:
                raise ValidationError(_("Invalid value for 'content_object': {}.").format(value))
            obj_type = ContentType.objects.get_for_model(obj, for_concrete_model=False)
            return queryset.filter(content_type_id=obj_type.id, object_id=obj.pk)

    class Meta:
        fields = (
            "role",
            "role__in",
            "role__contains",
            "role__icontains",
            "role__startswith",
            "content_object",
        )


class UserRoleFilter(NestedRoleFilter):
    """
    FilterSet for UserRole.
    """

    class Meta:
        model = UserRole
        fields = NestedRoleFilter.Meta.fields


class UserRoleViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for UserRole.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "roles"
    router_lookup = "user"
    parent_viewset = UserViewSet
    parent_lookup_kwargs = {"user_pk": "user__pk"}
    filterset_class = UserRoleFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = UserRoleSerializer
    queryset = UserRole.objects.all()
    ordering = ("-pulp_created",)


class GroupRoleFilter(NestedRoleFilter):
    """
    FilterSet for GroupRole.
    """

    class Meta:
        model = GroupRole
        fields = NestedRoleFilter.Meta.fields


class GroupRoleViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for GroupRole.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "roles"
    router_lookup = "group"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "group__pk"}
    filterset_class = GroupRoleFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = GroupRoleSerializer
    queryset = GroupRole.objects.all()
    ordering = ("-pulp_created",)



from drf_spectacular.utils import extend_schema
from rest_framework.views import APIView

from pulpcore.app.loggers import deprecation_logger
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import AsyncOperationResponseSerializer
from pulpcore.app.tasks import orphan_cleanup
from pulpcore.tasking.tasks import dispatch


class OrphansView(APIView):
    @extend_schema(
        description="DEPRECATED! Trigger an asynchronous task that deletes all "
        "orphaned content and artifacts. Use the `POST /pulp/api/v3/orphans/cleanup/` call "
        "instead.",
        summary="Delete orphans",
        responses={202: AsyncOperationResponseSerializer},
    )
    def delete(self, request, format=None):
        """
        Cleans up all the Content and Artifact orphans in the system
        """
        deprecation_logger.warning(
            "The `DELETE /pulp/api/v3/orphans/` call is deprecated. Use"
            "`POST /pulp/api/v3/orphans/cleanup/` instead."
        )

        task = dispatch(orphan_cleanup)

        return OperationPostponedResponse(task, request)



from gettext import gettext as _
from drf_spectacular.utils import extend_schema
from rest_framework.decorators import action
from rest_framework.serializers import ValidationError

from pulpcore.app import tasks
from pulpcore.app.models import Content, RepositoryVersion
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    RepositoryAddRemoveContentSerializer,
)
from pulpcore.app.viewsets import NamedModelViewSet
from pulpcore.tasking.tasks import dispatch


__all__ = ["ModifyRepositoryActionMixin"]


class ModifyRepositoryActionMixin:
    @extend_schema(
        description="Trigger an asynchronous task to create a new repository version.",
        summary="Modify Repository Content",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=RepositoryAddRemoveContentSerializer)
    def modify(self, request, pk):
        """
        Queues a task that creates a new RepositoryVersion by adding and removing content units
        """
        add_content_units = {}
        remove_content_units = {}

        repository = self.get_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if "base_version" in request.data:
            base_version_pk = self.get_resource(request.data["base_version"], RepositoryVersion).pk
        else:
            base_version_pk = None

        if "add_content_units" in request.data:
            for url in request.data["add_content_units"]:
                add_content_units[NamedModelViewSet.extract_pk(url)] = url

            content_units_pks = set(add_content_units.keys())
            existing_content_units = Content.objects.filter(pk__in=content_units_pks)
            existing_content_units.touch()

            self.verify_content_units(existing_content_units, add_content_units)

            add_content_units = list(add_content_units.keys())

        if "remove_content_units" in request.data:
            if "*" in request.data["remove_content_units"]:
                remove_content_units = ["*"]
            else:
                for url in request.data["remove_content_units"]:
                    remove_content_units[NamedModelViewSet.extract_pk(url)] = url
                content_units_pks = set(remove_content_units.keys())
                existing_content_units = Content.objects.filter(pk__in=content_units_pks)
                self.verify_content_units(existing_content_units, remove_content_units)
                remove_content_units = list(remove_content_units.keys())

        task = dispatch(
            tasks.repository.add_and_remove,
            exclusive_resources=[repository],
            kwargs={
                "repository_pk": pk,
                "base_version_pk": base_version_pk,
                "add_content_units": add_content_units,
                "remove_content_units": remove_content_units,
            },
        )
        return OperationPostponedResponse(task, request)

    def verify_content_units(self, content_units, all_content_units):
        """Verify referenced content units."""
        existing_content_units_pks = content_units.values_list("pk", flat=True)
        existing_content_units_pks = {str(pk) for pk in existing_content_units_pks}

        missing_pks = set(all_content_units.keys()) - existing_content_units_pks
        if missing_pks:
            missing_hrefs = [all_content_units[pk] for pk in missing_pks]
            raise ValidationError(
                _("Could not find the following content units: {}").format(missing_hrefs)
            )
